# -*- coding: utf-8 -*-
"""
tests for Tracer and utilities.
"""
import contextlib
import gc
import logging
import multiprocessing
import os
from os import getpid
import sys
import threading
from unittest.case import SkipTest
import weakref

import mock
import pytest
import six

import ddtrace
from ddtrace.constants import AUTO_KEEP
from ddtrace.constants import AUTO_REJECT
from ddtrace.constants import ENV_KEY
from ddtrace.constants import HOSTNAME_KEY
from ddtrace.constants import MANUAL_DROP_KEY
from ddtrace.constants import MANUAL_KEEP_KEY
from ddtrace.constants import ORIGIN_KEY
from ddtrace.constants import PID
from ddtrace.constants import SAMPLING_PRIORITY_KEY
from ddtrace.constants import USER_KEEP
from ddtrace.constants import USER_REJECT
from ddtrace.constants import VERSION_KEY
from ddtrace.context import Context
from ddtrace.contrib.trace_utils import set_user
from ddtrace.ext import user
from ddtrace.internal._encoding import MsgpackEncoderV03
from ddtrace.internal._encoding import MsgpackEncoderV05
from ddtrace.internal.writer import AgentWriter
from ddtrace.internal.writer import LogWriter
from ddtrace.settings import Config
from ddtrace.span import _is_top_level
from ddtrace.tracer import Tracer
from ddtrace.tracer import _has_aws_lambda_agent_extension
from ddtrace.tracer import _in_aws_lambda
from tests.subprocesstest import run_in_subprocess
from tests.utils import TracerTestCase
from tests.utils import override_global_config

from ..utils import override_env


class TracerTestCases(TracerTestCase):
    @pytest.fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    def test_tracer_vars(self):
        span = self.trace("a", service="s", resource="r", span_type="t")
        span.assert_matches(name="a", service="s", resource="r", span_type="t")
        # DEV: Finish to ensure we don't leak `service` between spans
        span.finish()

        span = self.trace("a")
        span.assert_matches(name="a", service=None, resource="a", span_type=None)
        span.finish()

    def test_tracer(self):
        def _mix():
            with self.trace("cake.mix"):
                pass

        def _bake():
            with self.trace("cake.bake"):
                pass

        def _make_cake():
            with self.trace("cake.make") as span:
                span.service = "baker"
                span.resource = "cake"
                _mix()
                _bake()

        # let's run it and make sure all is well.
        self.assert_has_no_spans()
        _make_cake()

        # Capture root's trace id to assert later
        root_trace_id = self.get_root_span().trace_id

        # Assert structure of this trace
        self.assert_structure(
            # Root span with 2 children
            dict(name="cake.make", resource="cake", service="baker", parent_id=None),
            (
                # Span with no children
                dict(name="cake.mix", resource="cake.mix", service="baker"),
                # Span with no children
                dict(name="cake.bake", resource="cake.bake", service="baker"),
            ),
        )

        # do it again and make sure it has new trace ids
        self.reset()
        _make_cake()
        self.assert_span_count(3)
        for s in self.spans:
            assert s.trace_id != root_trace_id

    def test_tracer_wrap(self):
        @self.tracer.wrap("decorated_function", service="s", resource="r", span_type="t")
        def f(tag_name, tag_value):
            # make sure we can still set tags
            span = self.tracer.current_span()
            span.set_tag(tag_name, tag_value)

        f("a", "b")

        self.assert_span_count(1)
        span = self.get_root_span()
        span.assert_matches(
            name="decorated_function",
            service="s",
            resource="r",
            span_type="t",
            meta=dict(a="b"),
        )

    def test_tracer_pid(self):
        with self.trace("root") as root_span:
            with self.trace("child") as child_span:
                pass

        # Root span should contain the pid of the current process
        root_span.assert_metrics({PID: getpid()}, exact=False)

        # Child span should not contain a pid tag
        child_span.assert_metrics(dict(), exact=True)

    def test_tracer_wrap_default_name(self):
        @self.tracer.wrap()
        def f():
            pass

        f()

        self.assert_structure(dict(name="tests.tracer.test_tracer.f"))

    def test_tracer_wrap_exception(self):
        @self.tracer.wrap()
        def f():
            raise Exception("bim")

        with self.assertRaises(Exception) as ex:
            f()

            self.assert_structure(
                dict(
                    name="tests.test_tracer.f",
                    error=1,
                    meta={
                        "error.msg": ex.message,
                        "error.type": ex.__class__.__name__,
                    },
                ),
            )

    def test_tracer_wrap_multiple_calls(self):
        @self.tracer.wrap()
        def f():
            pass

        f()
        f()

        self.assert_span_count(2)
        assert self.spans[0].span_id != self.spans[1].span_id

    def test_tracer_wrap_span_nesting_current_root_span(self):
        @self.tracer.wrap("inner")
        def inner():
            root_span = self.tracer.current_root_span()
            self.assertEqual(root_span.name, "outer")

        @self.tracer.wrap("outer")
        def outer():
            root_span = self.tracer.current_root_span()
            self.assertEqual(root_span.name, "outer")

            with self.trace("mid"):
                root_span = self.tracer.current_root_span()
                self.assertEqual(root_span.name, "outer")

                inner()

        outer()

    def test_tracer_wrap_span_nesting(self):
        @self.tracer.wrap("inner")
        def inner():
            pass

        @self.tracer.wrap("outer")
        def outer():
            with self.trace("mid"):
                inner()

        outer()

        self.assert_span_count(3)
        self.assert_structure(
            dict(name="outer"),
            ((dict(name="mid"), (dict(name="inner"),)),),
        )

    def test_tracer_wrap_class(self):
        class Foo(object):
            @staticmethod
            @self.tracer.wrap()
            def s():
                return 1

            @classmethod
            @self.tracer.wrap()
            def c(cls):
                return 2

            @self.tracer.wrap()
            def i(cls):
                return 3

        f = Foo()
        self.assertEqual(f.s(), 1)
        self.assertEqual(f.c(), 2)
        self.assertEqual(f.i(), 3)

        self.assert_span_count(3)
        self.spans[0].assert_matches(name="tests.tracer.test_tracer.s")
        self.spans[1].assert_matches(name="tests.tracer.test_tracer.c")
        self.spans[2].assert_matches(name="tests.tracer.test_tracer.i")

    def test_tracer_wrap_factory(self):
        def wrap_executor(tracer, fn, args, kwargs, span_name=None, service=None, resource=None, span_type=None):
            with tracer.trace("wrap.overwrite") as span:
                span.set_tag("args", args)
                span.set_tag("kwargs", kwargs)
                return fn(*args, **kwargs)

        @self.tracer.wrap()
        def wrapped_function(param, kw_param=None):
            self.assertEqual(42, param)
            self.assertEqual(42, kw_param)

        # set the custom wrap factory after the wrapper has been called
        self.tracer.configure(wrap_executor=wrap_executor)

        # call the function expecting that the custom tracing wrapper is used
        wrapped_function(42, kw_param=42)

        self.assert_span_count(1)
        self.spans[0].assert_matches(
            name="wrap.overwrite",
            meta=dict(args="(42,)", kwargs="{'kw_param': 42}"),
        )

    def test_tracer_wrap_factory_nested(self):
        def wrap_executor(tracer, fn, args, kwargs, span_name=None, service=None, resource=None, span_type=None):
            with tracer.trace("wrap.overwrite") as span:
                span.set_tag("args", args)
                span.set_tag("kwargs", kwargs)
                return fn(*args, **kwargs)

        @self.tracer.wrap()
        def wrapped_function(param, kw_param=None):
            self.assertEqual(42, param)
            self.assertEqual(42, kw_param)

        # set the custom wrap factory after the wrapper has been called
        self.tracer.configure(wrap_executor=wrap_executor)

        # call the function expecting that the custom tracing wrapper is used
        with self.trace("wrap.parent", service="webserver"):
            wrapped_function(42, kw_param=42)

        self.assert_structure(
            dict(name="wrap.parent", service="webserver"),
            (dict(name="wrap.overwrite", service="webserver", meta=dict(args="(42,)", kwargs="{'kw_param': 42}")),),
        )

    def test_tracer_disabled(self):
        self.tracer.enabled = True
        with self.trace("foo") as s:
            s.set_tag("a", "b")

        self.assert_has_spans()
        self.reset()

        self.tracer.enabled = False
        with self.trace("foo") as s:
            s.set_tag("a", "b")
        self.assert_has_no_spans()

    def test_unserializable_span_with_finish(self):
        try:
            import numpy as np
        except ImportError:
            raise SkipTest("numpy not installed")

        # a weird case where manually calling finish with an unserializable
        # span was causing an loop of serialization.
        with self.trace("parent") as span:
            span._metrics["as"] = np.int64(1)  # circumvent the data checks
            span.finish()

    def test_tracer_disabled_mem_leak(self):
        # ensure that if the tracer is disabled, we still remove things from the
        # span buffer upon finishing.
        self.tracer.enabled = False
        s1 = self.trace("foo")
        s1.finish()

        p1 = self.tracer.current_span()
        s2 = self.trace("bar")

        self.assertIsNone(s2._parent)
        s2.finish()
        self.assertIsNone(p1)

    def test_tracer_global_tags(self):
        s1 = self.trace("brie")
        s1.finish()
        self.assertIsNone(s1.get_tag("env"))
        self.assertIsNone(s1.get_tag("other"))

        self.tracer.set_tags({"env": "prod"})
        s2 = self.trace("camembert")
        s2.finish()
        self.assertEqual(s2.get_tag("env"), "prod")
        self.assertIsNone(s2.get_tag("other"))

        self.tracer.set_tags({"env": "staging", "other": "tag"})
        s3 = self.trace("gruyere")
        s3.finish()
        self.assertEqual(s3.get_tag("env"), "staging")
        self.assertEqual(s3.get_tag("other"), "tag")

    def test_global_context(self):
        # the tracer uses a global thread-local Context
        span = self.trace("fake_span")
        ctx = self.tracer.current_trace_context()
        assert ctx.trace_id == span.trace_id
        assert ctx.span_id == span.span_id

    def test_tracer_current_span(self):
        # the current span is in the local Context()
        span = self.trace("fake_span")
        assert self.tracer.current_span() == span
        span.finish()

        with self.trace("fake_span") as span:
            assert self.tracer.current_span() == span

    def test_tracer_current_span_missing_context(self):
        self.assertIsNone(self.tracer.current_span())

    def test_tracer_current_root_span_missing_context(self):
        self.assertIsNone(self.tracer.current_root_span())

    def test_default_provider_get(self):
        ctx = self.tracer.context_provider.active()
        assert ctx is None

    def test_default_provider_set(self):
        # The Context Provider can set the current active Context;
        # this could happen in distributed tracing
        ctx = Context(trace_id=42, span_id=100)
        self.tracer.context_provider.activate(ctx)
        span = self.trace("web.request")
        span.assert_matches(name="web.request", trace_id=42, parent_id=100)

    def test_start_span(self):
        # it should create a root Span
        span = self.tracer.start_span("web.request")
        assert span.name == "web.request"
        assert span.parent_id is None
        span.finish()
        spans = self.pop_spans()
        assert len(spans) == 1
        assert spans[0] is span

    def test_start_span_optional(self):
        # it should create a root Span with arguments
        with self.start_span("web.request", service="web", resource="/", span_type="http") as span:
            pass
        span.assert_matches(
            name="web.request",
            service="web",
            resource="/",
            span_type="http",
        )

    def test_start_span_service_default(self):
        span = self.start_span("")
        span.assert_matches(service=None)
        span.finish()

    def test_start_span_service_from_parent(self):
        with self.start_span("parent", service="mysvc") as parent:
            with self.start_span("child", child_of=parent) as child:
                pass
        child.assert_matches(
            name="child",
            service="mysvc",
        )

    def test_start_span_service_global_config(self):
        # When no service is provided a default
        with self.override_global_config(dict(service="mysvc")):
            with self.start_span("") as span:
                span.assert_matches(service="mysvc")

    def test_start_span_service_global_config_parent(self):
        # Parent should have precedence over global config
        with self.override_global_config(dict(service="mysvc")):
            with self.start_span("parent", service="parentsvc") as parent:
                with self.start_span("child", child_of=parent) as child:
                    pass
        child.assert_matches(
            name="child",
            service="parentsvc",
        )

    def test_start_child_span(self):
        # it should create a child Span for the given parent
        with self.start_span("web.request") as parent:
            assert self.tracer.current_span() is None
            with self.start_span("web.worker", child_of=parent) as child:
                assert self.tracer.current_span() is None

        parent.assert_matches(
            name="web.request",
            parent_id=None,
            _parent=None,
        )
        child.assert_matches(
            name="web.worker",
            parent_id=parent.span_id,
            _parent=parent,
        )

    def test_start_child_span_attributes(self):
        # it should create a child Span with parent's attributes
        with self.start_span("web.request", service="web", resource="/", span_type="http") as parent:
            with self.start_span("web.worker", child_of=parent) as child:
                child.assert_matches(name="web.worker", service="web")

    def test_start_child_from_context(self):
        # it should create a child span with a populated Context
        with self.start_span("web.request") as root:
            with self.start_span("web.worker", child_of=root.context) as child:
                pass
        child.assert_matches(
            name="web.worker",
            parent_id=root.span_id,
            trace_id=root.trace_id,
            _parent=None,
        )

    def test_adding_services(self):
        assert self.tracer._services == set()
        with self.start_span("root", service="one") as root:
            assert self.tracer._services == set(["one"])
            with self.start_span("child", service="two", child_of=root):
                pass
        assert self.tracer._services == set(["one", "two"])

    def test_configure_dogstatsd_url_host_port(self):
        tracer = Tracer()
        tracer.configure(dogstatsd_url="foo:1234")
        assert tracer._writer.dogstatsd.host == "foo"
        assert tracer._writer.dogstatsd.port == 1234

        tracer = Tracer()
        writer = AgentWriter("http://localhost:8126")
        tracer.configure(writer=writer, dogstatsd_url="foo:1234")
        assert tracer._writer.dogstatsd.host == "foo"
        assert tracer._writer.dogstatsd.port == 1234

    def test_configure_dogstatsd_url_socket(self):
        tracer = Tracer()
        tracer.configure(dogstatsd_url="unix:///foo.sock")
        assert tracer._writer.dogstatsd.host is None
        assert tracer._writer.dogstatsd.port is None
        assert tracer._writer.dogstatsd.socket_path == "/foo.sock"

        tracer = Tracer()
        writer = AgentWriter("http://localhost:8126")
        tracer.configure(writer=writer, dogstatsd_url="unix:///foo.sock")
        assert tracer._writer.dogstatsd.host is None
        assert tracer._writer.dogstatsd.port is None
        assert tracer._writer.dogstatsd.socket_path == "/foo.sock"

    def test_tracer_set_user(self):
        span = self.trace("fake_span")
        set_user(
            self.tracer,
            user_id="usr.id",
            email="usr.email",
            name="usr.name",
            session_id="usr.session_id",
            role="usr.role",
            scope="usr.scope",
        )
        assert span.get_tag(user.ID)
        assert span.get_tag(user.EMAIL)
        assert span.get_tag(user.SESSION_ID)
        assert span.get_tag(user.NAME)
        assert span.get_tag(user.ROLE)
        assert span.get_tag(user.SCOPE)
        assert span.context.dd_user_id is None

    def test_tracer_set_user_mandatory(self):
        span = self.trace("fake_span")
        set_user(
            self.tracer,
            user_id="usr.id",
        )
        span_keys = list(span.get_tags().keys())
        span_keys.sort()
        assert span_keys == ["runtime-id", "usr.id"]
        assert span.get_tag(user.ID)
        assert span.get_tag(user.EMAIL) is None
        assert span.get_tag(user.SESSION_ID) is None
        assert span.get_tag(user.NAME) is None
        assert span.get_tag(user.ROLE) is None
        assert span.get_tag(user.SCOPE) is None
        assert span.context.dd_user_id is None

    def test_tracer_set_user_warning_no_span(self):
        with self._caplog.at_level(logging.WARNING):
            set_user(
                self.tracer,
                user_id="usr.id",
            )
            assert "No root span in the current execution. Skipping set_user tags" in self._caplog.records[0].message

    def test_tracer_set_user_propagation(self):
        span = self.trace("fake_span")
        user_id_string = "usr.id"
        set_user(
            self.tracer,
            user_id=user_id_string,
            email="usr.email",
            name="usr.name",
            session_id="usr.session_id",
            role="usr.role",
            scope="usr.scope",
            propagate=True,
        )
        user_id = span.context._meta.get("_dd.p.usr.id")

        assert span.get_tag(user.ID) == user_id_string
        assert span.context.dd_user_id == user_id_string
        assert user_id == "dXNyLmlk"

    def test_tracer_set_user_propagation_empty(self):
        span = self.trace("fake_span")
        user_id_string = ""
        set_user(
            self.tracer,
            user_id=user_id_string,
            email="usr.email",
            name="usr.name",
            session_id="usr.session_id",
            role="usr.role",
            scope="usr.scope",
            propagate=True,
        )
        user_id = span.context._meta.get("_dd.p.usr.id")

        assert span.get_tag(user.ID) == user_id_string
        assert span.context.dd_user_id is None
        assert user_id == user_id_string

    @pytest.mark.skipif(sys.version_info < (3, 0, 0), reason="Python3 tests")
    def test_tracer_set_user_propagation_string_error_py3(self):
        span = self.trace("fake_span")
        user_id_string = "ユーザーID"
        set_user(
            self.tracer,
            user_id=user_id_string,
            email="usr.email",
            name="usr.name",
            session_id="usr.session_id",
            role="usr.role",
            scope="usr.scope",
            propagate=True,
        )
        user_id = span.context._meta.get("_dd.p.usr.id")

        assert span.get_tag(user.ID) == user_id_string
        assert span.context.dd_user_id == user_id_string
        assert user_id == "44Om44O844K244O8SUQ="

    @pytest.mark.skipif(sys.version_info >= (3, 0, 0), reason="Python tests")
    def test_tracer_set_user_propagation_string_error_py2(self):
        span = self.trace("fake_span")
        user_id_string = "ユーザーID"
        set_user(
            self.tracer,
            user_id="ユーザーID",
            email="usr.email",
            name="usr.name",
            session_id="usr.session_id",
            role="usr.role",
            scope="usr.scope",
            propagate=True,
        )
        user_id = span.context._meta.get("_dd.p.usr.id")

        assert span.get_tag(user.ID) is None
        assert span.context.dd_user_id == user_id_string
        assert user_id == "44Om44O844K244O8SUQ="


def test_tracer_url():
    t = ddtrace.Tracer()
    assert t._writer.agent_url == "http://localhost:8126"

    t = ddtrace.Tracer(url="http://foobar:12")
    assert t._writer.agent_url == "http://foobar:12"

    t = ddtrace.Tracer(url="unix:///foobar")
    assert t._writer.agent_url == "unix:///foobar"

    t = ddtrace.Tracer(url="http://localhost")
    assert t._writer.agent_url == "http://localhost"

    t = ddtrace.Tracer(url="https://localhost")
    assert t._writer.agent_url == "https://localhost"

    with pytest.raises(ValueError) as e:
        ddtrace.Tracer(url="foo://foobar:12")
    assert (
        str(e.value) == "Unsupported protocol 'foo' in Agent URL 'foo://foobar:12'. Must be one of: http, https, unix"
    )


def test_tracer_shutdown_no_timeout():
    t = ddtrace.Tracer()

    with mock.patch.object(AgentWriter, "stop") as mock_stop:
        with mock.patch.object(AgentWriter, "join") as mock_join:
            t.shutdown()

    mock_stop.assert_called()
    mock_join.assert_not_called()


def test_tracer_configure_writer_stop_unstarted():
    t = ddtrace.Tracer()
    t._writer = mock.Mock(wraps=t._writer)
    orig_writer = t._writer

    # Stop should be called when replacing the writer.
    t.configure(hostname="localhost", port=8126)
    assert orig_writer.stop.called


def test_tracer_configure_writer_stop_started():
    t = ddtrace.Tracer()
    t._writer = mock.Mock(wraps=t._writer)
    orig_writer = t._writer

    # Do a write to start the writer
    with t.trace("something"):
        pass

    t.configure(hostname="localhost", port=8126)
    orig_writer.stop.assert_called_once_with()


def test_tracer_shutdown_timeout():
    t = ddtrace.Tracer()

    with mock.patch.object(AgentWriter, "stop") as mock_stop:
        with t.trace("something"):
            pass

        t.shutdown(timeout=2)
    mock_stop.assert_called_once_with(2)


def test_tracer_shutdown():
    t = ddtrace.Tracer()
    t.shutdown()

    with mock.patch.object(AgentWriter, "write") as mock_write:
        with t.trace("something"):
            pass

    mock_write.assert_not_called()


def test_tracer_shutdown_warning():
    t = ddtrace.Tracer()
    t.shutdown()

    with mock.patch.object(logging.Logger, "warning") as mock_logger:
        with t.trace("something"):
            pass

    mock_logger.assert_has_calls(
        [
            mock.call("Spans started after the tracer has been shut down will not be sent to the Datadog Agent."),
        ]
    )


def test_tracer_dogstatsd_url():
    t = ddtrace.Tracer()
    assert t._writer.dogstatsd.host == "localhost"
    assert t._writer.dogstatsd.port == 8125

    t = ddtrace.Tracer(dogstatsd_url="foobar:12")
    assert t._writer.dogstatsd.host == "foobar"
    assert t._writer.dogstatsd.port == 12

    t = ddtrace.Tracer(dogstatsd_url="udp://foobar:12")
    assert t._writer.dogstatsd.host == "foobar"
    assert t._writer.dogstatsd.port == 12

    t = ddtrace.Tracer(dogstatsd_url="/var/run/statsd.sock")
    assert t._writer.dogstatsd.socket_path == "/var/run/statsd.sock"

    t = ddtrace.Tracer(dogstatsd_url="unix:///var/run/statsd.sock")
    assert t._writer.dogstatsd.socket_path == "/var/run/statsd.sock"

    with pytest.raises(ValueError) as e:
        t = ddtrace.Tracer(dogstatsd_url="foo://foobar:12")
        assert str(e) == "Unknown url format for `foo://foobar:12`"


def test_tracer_fork():
    t = ddtrace.Tracer()
    original_pid = t._pid
    original_writer = t._writer

    @contextlib.contextmanager
    def capture_failures(errors):
        try:
            yield
        except AssertionError as e:
            errors.put(e)

    def task(t, errors):
        # Start a new span to trigger process checking
        with t.trace("test", service="test"):

            # Assert we recreated the writer and have a new queue
            with capture_failures(errors):
                assert t._pid != original_pid
                assert t.writer is not original_writer
                assert t.writer._encoder is not original_writer._encoder

        # Assert the trace got written into the correct queue
        assert len(original_writer._encoder) == 0
        assert len(t.writer._encoder) == 1

    # Assert tracer in a new process correctly recreates the writer
    errors = multiprocessing.Queue()
    p = multiprocessing.Process(target=task, args=(t, errors))
    try:
        p.start()
    finally:
        p.join(timeout=2)

    assert errors.empty(), errors.get()

    # Ensure writing into the tracer in this process still works as expected
    with t.trace("test", service="test"):
        assert t._pid == original_pid
        assert t._writer == original_writer
        assert t._writer._encoder == original_writer._encoder

    # Assert the trace got written into the correct queue
    assert len(original_writer._encoder) == 1
    assert len(t._writer._encoder) == 1


def test_tracer_with_version():
    t = ddtrace.Tracer()

    # With global `config.version` defined
    with override_global_config(dict(version="1.2.3")):
        with t.trace("test.span") as span:
            assert span.get_tag(VERSION_KEY) == "1.2.3"

            # override manually
            span.set_tag(VERSION_KEY, "4.5.6")
            assert span.get_tag(VERSION_KEY) == "4.5.6"

    # With no `config.version` defined
    with t.trace("test.span") as span:
        assert span.get_tag(VERSION_KEY) is None

        # explicitly set in the span
        span.set_tag(VERSION_KEY, "1.2.3")
        assert span.get_tag(VERSION_KEY) == "1.2.3"

    # With global tags set
    t.set_tags({VERSION_KEY: "tags.version"})
    with override_global_config(dict(version="config.version")):
        with t.trace("test.span") as span:
            assert span.get_tag(VERSION_KEY) == "config.version"


def test_tracer_with_env():
    t = ddtrace.Tracer()

    # With global `config.env` defined
    with override_global_config(dict(env="prod")):
        with t.trace("test.span") as span:
            assert span.get_tag(ENV_KEY) == "prod"

            # override manually
            span.set_tag(ENV_KEY, "prod-staging")
            assert span.get_tag(ENV_KEY) == "prod-staging"

    # With no `config.env` defined
    with t.trace("test.span") as span:
        assert span.get_tag(ENV_KEY) is None

        # explicitly set in the span
        span.set_tag(ENV_KEY, "prod-staging")
        assert span.get_tag(ENV_KEY) == "prod-staging"

    # With global tags set
    t.set_tags({ENV_KEY: "tags.env"})
    with override_global_config(dict(env="config.env")):
        with t.trace("test.span") as span:
            assert span.get_tag(ENV_KEY) == "config.env"


class EnvTracerTestCase(TracerTestCase):
    """Tracer test cases requiring environment variables."""

    @run_in_subprocess(env_overrides=dict(DD_SERVICE="mysvc"))
    def test_service_name_env(self):
        with self.start_span("") as span:
            pass
        span.assert_matches(
            service="mysvc",
        )

    @run_in_subprocess(env_overrides=dict(DD_SERVICE="mysvc"))
    def test_service_name_env_global_config(self):
        # Global config should have higher precedence than the environment variable
        with self.override_global_config(dict(service="overridesvc")):
            with self.start_span("") as span:
                pass
        span.assert_matches(
            service="overridesvc",
        )

    @run_in_subprocess(env_overrides=dict(DD_VERSION="0.1.2"))
    def test_version_no_global_service(self):
        # Version should be set if no service name is present
        with self.trace("") as span:
            span.assert_matches(
                meta={
                    VERSION_KEY: "0.1.2",
                },
            )

        # The version will not be tagged if the service is not globally
        # configured.
        with self.trace("root", service="rootsvc") as root:
            assert VERSION_KEY not in root.get_tags()
            with self.trace("child") as span:
                assert VERSION_KEY not in span.get_tags()

    @run_in_subprocess(env_overrides=dict(DD_SERVICE="django", DD_VERSION="0.1.2"))
    def test_version_service(self):
        # Fleshed out example of service and version tagging

        # Our app is called django, we provide DD_SERVICE=django and DD_VERSION=0.1.2

        with self.trace("django.request") as root:
            # Root span should be tagged
            assert root.service == "django"
            assert VERSION_KEY in root.get_tags() and root.get_tag(VERSION_KEY) == "0.1.2"

            # Child spans should be tagged
            with self.trace("") as child1:
                assert child1.service == "django"
                assert VERSION_KEY in child1.get_tags() and child1.get_tag(VERSION_KEY) == "0.1.2"

            # Version should not be applied to spans of a service that isn't user-defined
            with self.trace("mysql.query", service="mysql") as span:
                assert VERSION_KEY not in span.get_tags()
                # Child should also not have a version
                with self.trace("") as child2:
                    assert child2.service == "mysql"
                    assert VERSION_KEY not in child2.get_tags()

    @run_in_subprocess(env_overrides=dict(AWS_LAMBDA_FUNCTION_NAME="my-func"))
    def test_detect_agentless_env_with_lambda(self):
        assert _in_aws_lambda()
        assert not _has_aws_lambda_agent_extension()
        tracer = Tracer()
        assert isinstance(tracer._writer, LogWriter)
        tracer.configure(enabled=True)
        assert isinstance(tracer._writer, LogWriter)

    @run_in_subprocess(env_overrides=dict(AWS_LAMBDA_FUNCTION_NAME="my-func"))
    def test_detect_agent_config_with_lambda_extension(self):
        def mock_os_path_exists(path):
            return path == "/opt/extensions/datadog-agent"

        assert _in_aws_lambda()

        with mock.patch("os.path.exists", side_effect=mock_os_path_exists):
            assert _has_aws_lambda_agent_extension()

            tracer = Tracer()
            assert isinstance(tracer._writer, AgentWriter)
            assert tracer._writer._sync_mode

            tracer.configure(enabled=False)
            assert isinstance(tracer._writer, AgentWriter)
            assert tracer._writer._sync_mode

    @run_in_subprocess(env_overrides=dict(AWS_LAMBDA_FUNCTION_NAME="my-func", DD_AGENT_HOST="localhost"))
    def test_detect_agent_config(self):
        tracer = Tracer()
        assert isinstance(tracer._writer, AgentWriter)

    @run_in_subprocess(env_overrides=dict(DD_TAGS="key1:value1,key2:value2"))
    def test_dd_tags(self):
        assert self.tracer._tags.get("key1") == "value1"
        assert self.tracer._tags.get("key2") == "value2"

    @run_in_subprocess(env_overrides=dict(DD_TAGS="key1:value1,key2:value2,key3"))
    def test_dd_tags_invalid(self):
        assert self.tracer._tags.get("key1")
        assert self.tracer._tags.get("key2")
        assert self.tracer._tags.get("key3") is None

    @run_in_subprocess(env_overrides=dict(DD_TAGS="service:mysvc,env:myenv,version:myvers"))
    def test_tags_from_DD_TAGS(self):
        t = ddtrace.Tracer()
        with t.trace("test") as s:
            assert s.service == "mysvc"
            assert s.get_tag("env") == "myenv"
            assert s.get_tag("version") == "myvers"

    @run_in_subprocess(
        env_overrides=dict(
            DD_TAGS="service:s,env:e,version:v",
            DD_ENV="env",
            DD_SERVICE="svc",
            DD_VERSION="0.123",
        )
    )
    def test_tags_from_DD_TAGS_precedence(self):
        t = ddtrace.Tracer()
        with t.trace("test") as s:
            assert s.service == "svc"
            assert s.get_tag("env") == "env"
            assert s.get_tag("version") == "0.123"

    @run_in_subprocess(env_overrides=dict(DD_TAGS="service:mysvc,env:myenv,version:myvers"))
    def test_tags_from_DD_TAGS_override(self):
        t = ddtrace.Tracer()
        ddtrace.config.env = "env"
        ddtrace.config.service = "service"
        ddtrace.config.version = "0.123"
        with t.trace("test") as s:
            assert s.service == "service"
            assert s.get_tag("env") == "env"
            assert s.get_tag("version") == "0.123"


def test_tracer_set_runtime_tags():
    t = ddtrace.Tracer()
    with t.start_span("foobar") as span:
        pass

    assert len(span.get_tag("runtime-id"))

    t2 = ddtrace.Tracer()
    with t2.start_span("foobaz") as span2:
        pass

    assert span.get_tag("runtime-id") == span2.get_tag("runtime-id")


def _test_tracer_runtime_tags_fork_task(tracer, q):
    span = tracer.start_span("foobaz")
    q.put(span.get_tag("runtime-id"))
    span.finish()


def test_tracer_runtime_tags_fork():
    tracer = ddtrace.Tracer()

    span = tracer.start_span("foobar")
    span.finish()

    q = multiprocessing.Queue()
    p = multiprocessing.Process(target=_test_tracer_runtime_tags_fork_task, args=(tracer, q))
    p.start()
    p.join()

    children_tag = q.get()
    assert children_tag != span.get_tag("runtime-id")


def test_tracer_runtime_tags_cross_execution(tracer):
    ctx = Context(trace_id=12, span_id=21)
    tracer.context_provider.activate(ctx)
    with tracer.trace("span") as span:
        pass
    assert span.get_tag("runtime-id") is not None
    assert span.get_metric(PID) is not None


def test_start_span_hooks():
    t = ddtrace.Tracer()

    result = {}

    @t.on_start_span
    def store_span(span):
        result["span"] = span

    span = t.start_span("hello")

    assert span == result["span"]
    span.finish()


def test_deregister_start_span_hooks():
    t = ddtrace.Tracer()

    result = {}

    @t.on_start_span
    def store_span(span):
        result["span"] = span

    t.deregister_on_start_span(store_span)

    with t.start_span("hello"):
        pass

    assert result == {}


def test_enable(monkeypatch):
    t1 = ddtrace.Tracer()
    assert t1.enabled

    monkeypatch.setenv("DD_TRACE_ENABLED", "false")
    t2 = ddtrace.Tracer()
    assert not t2.enabled


def test_runtime_id_parent_only():
    tracer = ddtrace.Tracer()

    # Parent spans should have runtime-id
    s = tracer.trace("test")
    rtid = s.get_tag("runtime-id")
    assert isinstance(rtid, six.string_types)

    # Child spans should not
    s2 = tracer.trace("test2")
    assert s2.get_tag("runtime-id") is None
    s2.finish()
    s.finish()

    # Parent spans should have runtime-id
    s = tracer.trace("test")
    s.finish()
    rtid = s.get_tag("runtime-id")
    assert isinstance(rtid, six.string_types)


def test_runtime_id_fork():
    tracer = ddtrace.Tracer()

    s = tracer.trace("test")
    s.finish()

    rtid = s.get_tag("runtime-id")
    assert isinstance(rtid, six.string_types)

    pid = os.fork()

    if pid == 0:
        # child
        s = tracer.trace("test")
        s.finish()

        rtid_child = s.get_tag("runtime-id")
        assert isinstance(rtid_child, six.string_types)
        assert rtid != rtid_child
        os._exit(12)

    _, status = os.waitpid(pid, 0)
    exit_code = os.WEXITSTATUS(status)
    assert exit_code == 12


def test_multiple_tracer_ctx():
    t1 = ddtrace.Tracer()
    t2 = ddtrace.Tracer()

    with t1.trace("") as s1:
        with t2.trace("") as s2:
            pass

    assert s2.parent_id == s1.span_id
    assert s2.trace_id == s1.trace_id


def test_filters(tracer, test_spans):
    class FilterAll(object):
        def process_trace(self, trace):
            return None

    tracer.configure(
        settings={
            "FILTERS": [FilterAll()],
        }
    )

    with tracer.trace("root"):
        with tracer.trace("child"):
            pass

    spans = test_spans.pop()
    assert len(spans) == 0

    class FilterMutate(object):
        def __init__(self, key, value):
            self.key = key
            self.value = value

        def process_trace(self, trace):
            for s in trace:
                s.set_tag(self.key, self.value)
            return trace

    tracer.configure(
        settings={
            "FILTERS": [FilterMutate("boop", "beep")],
        }
    )

    with tracer.trace("root"):
        with tracer.trace("child"):
            pass

    spans = test_spans.pop()
    assert len(spans) == 2
    s1, s2 = spans
    assert s1.get_tag("boop") == "beep"
    assert s2.get_tag("boop") == "beep"

    # Test multiple filters
    tracer.configure(
        settings={
            "FILTERS": [FilterMutate("boop", "beep"), FilterMutate("mats", "sundin")],
        }
    )

    with tracer.trace("root"):
        with tracer.trace("child"):
            pass

    spans = test_spans.pop()
    assert len(spans) == 2
    for s in spans:
        assert s.get_tag("boop") == "beep"
        assert s.get_tag("mats") == "sundin"

    class FilterBroken(object):
        def process_trace(self, trace):
            _ = 1 / 0

    tracer.configure(
        settings={
            "FILTERS": [FilterBroken()],
        }
    )

    with tracer.trace("root"):
        with tracer.trace("child"):
            pass

    spans = test_spans.pop()
    assert len(spans) == 2

    tracer.configure(
        settings={
            "FILTERS": [FilterMutate("boop", "beep"), FilterBroken()],
        }
    )
    with tracer.trace("root"):
        with tracer.trace("child"):
            pass

    spans = test_spans.pop()
    assert len(spans) == 2
    for s in spans:
        assert s.get_tag("boop") == "beep"


def test_early_exit(tracer, test_spans):
    s1 = tracer.trace("1")
    s2 = tracer.trace("2")
    with mock.patch.object(logging.Logger, "debug") as mock_logger:
        s1.finish()
        s2.finish()

    calls = [
        mock.call("span %r closing after its parent %r, this is an error when not using async", s2, s1),
    ]
    mock_logger.assert_has_calls(calls)
    assert s1.parent_id is None
    assert s2.parent_id is s1.span_id

    traces = test_spans.pop_traces()
    assert len(traces) == 1
    assert len(traces[0]) == 2

    s1 = tracer.trace("1-1")
    s1.finish()
    assert s1.parent_id is None

    s1 = tracer.trace("1-2")
    s1.finish()
    assert s1.parent_id is None


class TestPartialFlush(TracerTestCase):
    @TracerTestCase.run_in_subprocess(
        env_overrides=dict(DD_TRACE_PARTIAL_FLUSH_ENABLED="true", DD_TRACE_PARTIAL_FLUSH_MIN_SPANS="5")
    )
    def test_partial_flush(self):
        self._test_partial_flush()

    @TracerTestCase.run_in_subprocess(
        env_overrides=dict(DD_TRACE_PARTIAL_FLUSH_ENABLED="true", DD_TRACE_PARTIAL_FLUSH_MIN_SPANS="1")
    )
    def test_partial_flush_too_many(self):
        root = self.tracer.trace("root")
        for i in range(5):
            self.tracer.trace("child%s" % i).finish()

        traces = self.pop_traces()
        assert len(traces) == 5
        for t in traces:
            assert len(t) == 1
        assert [t[0].name for t in traces] == ["child0", "child1", "child2", "child3", "child4"]
        for t in traces:
            assert t[0].parent_id == root.span_id

        root.finish()
        traces = self.pop_traces()
        assert len(traces) == 1
        assert traces[0][0].name == "root"

    @TracerTestCase.run_in_subprocess(
        env_overrides=dict(DD_TRACE_PARTIAL_FLUSH_ENABLED="true", DD_TRACE_PARTIAL_FLUSH_MIN_SPANS="6")
    )
    def test_partial_flush_too_few(self):
        root = self.tracer.trace("root")
        for i in range(5):
            self.tracer.trace("child%s" % i).finish()

        traces = self.pop_traces()
        assert len(traces) == 0
        root.finish()
        traces = self.pop_traces()
        assert len(traces) == 1
        assert [s.name for s in traces[0]] == ["root", "child0", "child1", "child2", "child3", "child4"]

    def test_partial_flush_configure(self):
        self.tracer.configure(partial_flush_enabled=True, partial_flush_min_spans=5)
        self.test_partial_flush()

    def test_partial_flush_too_many_configure(self):
        self.tracer.configure(partial_flush_enabled=True, partial_flush_min_spans=1)
        self.test_partial_flush_too_many()

    def test_partial_flush_too_few_configure(self):
        self.tracer.configure(partial_flush_enabled=True, partial_flush_min_spans=6)
        self.test_partial_flush_too_few()

    @TracerTestCase.run_in_subprocess(
        env_overrides=dict(DD_TRACE_PARTIAL_FLUSH_ENABLED="false", DD_TRACE_PARTIAL_FLUSH_MIN_SPANS="6")
    )
    def test_partial_flush_configure_precedence(self):
        self.tracer.configure(partial_flush_enabled=True, partial_flush_min_spans=5)
        self.test_partial_flush()

    def _test_partial_flush(self):
        root = self.tracer.trace("root")
        for i in range(5):
            self.tracer.trace("child%s" % i).finish()

        traces = self.pop_traces()
        assert len(traces) == 1
        assert len(traces[0]) == 5
        assert [s.name for s in traces[0]] == ["child0", "child1", "child2", "child3", "child4"]

        root.finish()
        traces = self.pop_traces()
        assert len(traces) == 1
        assert len(traces[0]) == 1
        assert traces[0][0].name == "root"


def test_unicode_config_vals():
    t = ddtrace.Tracer()

    with override_global_config(dict(version=u"😇", env=u"😇")):
        with t.trace("1"):
            pass
    t.shutdown()


def test_ctx(tracer, test_spans):
    with tracer.trace("test") as s1:
        assert tracer.current_span() == s1
        assert tracer.current_root_span() == s1
        assert tracer.current_trace_context().trace_id == s1.trace_id
        assert tracer.current_trace_context().span_id == s1.span_id

        with tracer.trace("test2") as s2:
            assert tracer.current_span() == s2
            assert tracer.current_root_span() == s1
            assert tracer.current_trace_context().trace_id == s1.trace_id
            assert tracer.current_trace_context().span_id == s2.span_id

            with tracer.trace("test3") as s3:
                assert tracer.current_span() == s3
                assert tracer.current_root_span() == s1
                assert tracer.current_trace_context().trace_id == s1.trace_id
                assert tracer.current_trace_context().span_id == s3.span_id

            assert tracer.current_trace_context().trace_id == s1.trace_id
            assert tracer.current_trace_context().span_id == s2.span_id

        with tracer.trace("test4") as s4:
            assert tracer.current_span() == s4
            assert tracer.current_root_span() == s1
            assert tracer.current_trace_context().trace_id == s1.trace_id
            assert tracer.current_trace_context().span_id == s4.span_id

        assert tracer.current_span() == s1
        assert tracer.current_root_span() == s1

    assert tracer.current_span() is None
    assert tracer.current_root_span() is None
    assert s1.parent_id is None
    assert s2.parent_id == s1.span_id
    assert s3.parent_id == s2.span_id
    assert s4.parent_id == s1.span_id
    assert s1.trace_id == s2.trace_id == s3.trace_id == s4.trace_id
    assert s1.get_metric(SAMPLING_PRIORITY_KEY) == 1
    assert s2.get_metric(SAMPLING_PRIORITY_KEY) is None
    assert ORIGIN_KEY not in s1.get_tags()

    t = test_spans.pop_traces()
    assert len(t) == 1
    assert len(t[0]) == 4
    _s1, _s2, _s3, _s4 = t[0]
    assert s1 == _s1
    assert s2 == _s2
    assert s3 == _s3
    assert s4 == _s4

    with tracer.trace("s") as s:
        assert s.parent_id is None
        assert s.trace_id != s1.trace_id


def test_multithreaded(tracer, test_spans):
    def target():
        with tracer.trace("s1"):
            with tracer.trace("s2"):
                pass
            with tracer.trace("s3"):
                pass

    for i in range(1000):
        ts = [threading.Thread(target=target) for _ in range(10)]
        for t in ts:
            t.start()

        for t in ts:
            t.join()

        traces = test_spans.pop_traces()
        assert len(traces) == 10

        for trace in traces:
            assert len(trace) == 3


def test_ctx_distributed(tracer, test_spans):
    # Test activating an invalid context.
    ctx = Context(span_id=None, trace_id=None)
    tracer.context_provider.activate(ctx)
    assert tracer.current_span() is None

    with tracer.trace("test") as s1:
        assert tracer.current_span() == s1
        assert tracer.current_root_span() == s1
        assert tracer.current_trace_context().trace_id == s1.trace_id
        assert tracer.current_trace_context().span_id == s1.span_id
        assert s1.parent_id is None

    trace = test_spans.pop_traces()
    assert len(trace) == 1

    # Test activating a valid context.
    ctx = Context(span_id=1234, trace_id=4321, sampling_priority=2, dd_origin="somewhere")
    tracer.context_provider.activate(ctx)
    assert tracer.current_span() is None
    assert (
        tracer.current_trace_context()
        == tracer.context_provider.active()
        == Context(span_id=1234, trace_id=4321, sampling_priority=2, dd_origin="somewhere")
    )

    with tracer.trace("test2") as s2:
        assert tracer.current_span() == s2
        assert tracer.current_root_span() == s2
        assert tracer.current_trace_context().trace_id == s2.trace_id == 4321
        assert tracer.current_trace_context().span_id == s2.span_id
        assert s2.parent_id == 1234

    trace = test_spans.pop_traces()
    assert len(trace) == 1
    assert s2.get_metric(SAMPLING_PRIORITY_KEY) == 2
    assert s2.get_tag(ORIGIN_KEY) == "somewhere"


def test_manual_keep(tracer, test_spans):
    # On a root span
    with tracer.trace("asdf") as s:
        s.set_tag(MANUAL_KEEP_KEY)
    spans = test_spans.pop()
    assert spans[0].get_metric(SAMPLING_PRIORITY_KEY) is USER_KEEP

    # On a child span
    with tracer.trace("asdf"):
        with tracer.trace("child") as s:
            s.set_tag(MANUAL_KEEP_KEY)
    spans = test_spans.pop()
    assert spans[0].get_metric(SAMPLING_PRIORITY_KEY) is USER_KEEP


def test_manual_keep_then_drop(tracer, test_spans):
    # Test changing the value before finish.
    with tracer.trace("asdf") as root:
        with tracer.trace("child") as child:
            child.set_tag(MANUAL_KEEP_KEY)
        root.set_tag(MANUAL_DROP_KEY)
    spans = test_spans.pop()
    assert spans[0].get_metric(SAMPLING_PRIORITY_KEY) is USER_REJECT


def test_manual_drop(tracer, test_spans):
    # On a root span
    with tracer.trace("asdf") as s:
        s.set_tag(MANUAL_DROP_KEY)
    spans = test_spans.pop()
    assert spans[0].get_metric(SAMPLING_PRIORITY_KEY) is USER_REJECT

    # On a child span
    with tracer.trace("asdf"):
        with tracer.trace("child") as s:
            s.set_tag(MANUAL_DROP_KEY)
    spans = test_spans.pop()
    assert spans[0].get_metric(SAMPLING_PRIORITY_KEY) is USER_REJECT


@mock.patch("ddtrace.internal.hostname.get_hostname")
def test_get_report_hostname_enabled(get_hostname, tracer, test_spans):
    get_hostname.return_value = "test-hostname"
    with override_global_config(dict(report_hostname=True)):
        with tracer.trace("span"):
            with tracer.trace("child"):
                pass

    spans = test_spans.pop()
    root = spans[0]
    child = spans[1]
    assert root.get_tag(HOSTNAME_KEY) == "test-hostname"
    assert child.get_tag(HOSTNAME_KEY) is None


@mock.patch("ddtrace.internal.hostname.get_hostname")
def test_get_report_hostname_disabled(get_hostname, tracer, test_spans):
    get_hostname.return_value = "test-hostname"
    with override_global_config(dict(report_hostname=False)):
        with tracer.trace("span"):
            with tracer.trace("child"):
                pass

    spans = test_spans.pop()
    root = spans[0]
    child = spans[1]
    assert root.get_tag(HOSTNAME_KEY) is None
    assert child.get_tag(HOSTNAME_KEY) is None


@mock.patch("ddtrace.internal.hostname.get_hostname")
def test_get_report_hostname_default(get_hostname, tracer, test_spans):
    get_hostname.return_value = "test-hostname"
    with override_global_config(dict(report_hostname=False)):
        with tracer.trace("span"):
            with tracer.trace("child"):
                pass

    spans = test_spans.pop()
    root = spans[0]
    child = spans[1]
    assert root.get_tag(HOSTNAME_KEY) is None
    assert child.get_tag(HOSTNAME_KEY) is None


def test_non_active_span(tracer, test_spans):
    with tracer.start_span("test", activate=False):
        assert tracer.current_span() is None
        assert tracer.current_root_span() is None
    assert tracer.current_span() is None
    assert tracer.current_root_span() is None
    traces = test_spans.pop_traces()
    assert len(traces) == 1
    assert len(traces[0]) == 1

    with tracer.start_span("test1", activate=False):
        with tracer.start_span("test2", activate=False):
            assert tracer.current_span() is None
            assert tracer.current_root_span() is None
    assert tracer.current_span() is None
    assert tracer.current_root_span() is None
    traces = test_spans.pop_traces()
    assert len(traces) == 2

    with tracer.start_span("active", activate=True) as active:
        with tracer.start_span("non active", child_of=active, activate=False):
            assert tracer.context_provider.active() is active
            assert tracer.current_root_span() is active
        assert tracer.context_provider.active() is active
        assert tracer.current_root_span() is active
    traces = test_spans.pop_traces()
    assert len(traces) == 1
    assert len(traces[0]) == 2


def test_service_mapping():
    @contextlib.contextmanager
    def override_service_mapping(service_mapping):
        with override_env(dict(DD_SERVICE_MAPPING=service_mapping)):
            assert ddtrace.config.service_mapping == {}
            ddtrace.config.service_mapping = Config().service_mapping
            yield
            ddtrace.config.service_mapping = {}

    # Test single mapping
    with override_service_mapping("foo:bar"), ddtrace.Tracer().trace("renaming", service="foo") as span:
        assert span.service == "bar"

    # Test multiple mappings
    with override_service_mapping("foo:bar,sna:fu"), ddtrace.Tracer().trace("renaming", service="sna") as span:
        assert span.service == "fu"

    # Test colliding mappings
    with override_service_mapping("foo:bar,foo:foobar"), ddtrace.Tracer().trace("renaming", service="foo") as span:
        assert span.service == "foobar"

    # Test invalid service mapping
    with override_service_mapping("foo;bar,sna:fu"):
        with ddtrace.Tracer().trace("passthru", service="foo") as _:
            assert _.service == "foo"
        with ddtrace.Tracer().trace("renaming", "sna") as _:
            assert _.service == "fu"


def test_configure_url_partial():
    tracer = ddtrace.Tracer()
    tracer.configure(hostname="abc")
    assert tracer._writer.agent_url == "http://abc:8126"
    tracer.configure(port=123)
    assert tracer._writer.agent_url == "http://abc:123"

    tracer = ddtrace.Tracer(url="http://abc")
    assert tracer._writer.agent_url == "http://abc"
    tracer.configure(port=123)
    assert tracer._writer.agent_url == "http://abc:123"
    tracer.configure(port=431)
    assert tracer._writer.agent_url == "http://abc:431"


def test_bad_agent_url(monkeypatch):
    with pytest.raises(ValueError):
        Tracer(url="bad://localhost:8126")

    monkeypatch.setenv("DD_TRACE_AGENT_URL", "bad://localhost:1234")
    with pytest.raises(ValueError) as e:
        Tracer()
    assert (
        str(e.value)
        == "Unsupported protocol 'bad' in Agent URL 'bad://localhost:1234'. Must be one of: http, https, unix"
    )

    monkeypatch.setenv("DD_TRACE_AGENT_URL", "unix://")
    with pytest.raises(ValueError) as e:
        Tracer()
    assert str(e.value) == "Invalid file path in Agent URL 'unix://'"

    monkeypatch.setenv("DD_TRACE_AGENT_URL", "http://")
    with pytest.raises(ValueError) as e:
        Tracer()
    assert str(e.value) == "Invalid hostname in Agent URL 'http://'"


def test_context_priority(tracer, test_spans):
    """Assigning a sampling_priority should not affect if the trace is sent to the agent"""
    for p in [USER_REJECT, AUTO_REJECT, AUTO_KEEP, USER_KEEP, None, 999]:
        with tracer.trace("span_%s" % p) as span:
            span.context.sampling_priority = p

        # Spans should always be written regardless of sampling priority since
        # the agent needs to know the sampling decision.
        spans = test_spans.pop()
        assert len(spans) == 1, "trace should be sampled"
        if p in [USER_REJECT, AUTO_REJECT, AUTO_KEEP, USER_KEEP]:
            assert spans[0].get_metric(SAMPLING_PRIORITY_KEY) == p


def test_spans_sampled_out(tracer, test_spans):
    with tracer.trace("root") as span:
        span.sampled = False
        with tracer.trace("child") as span:
            span.sampled = False
        with tracer.trace("child") as span:
            span.sampled = False

    spans = test_spans.pop()
    assert len(spans) == 0


def test_spans_sampled_one(tracer, test_spans):
    with tracer.trace("root") as span:
        span.sampled = False
        with tracer.trace("child") as span:
            span.sampled = False
        with tracer.trace("child") as span:
            span.sampled = True

    spans = test_spans.pop()
    assert len(spans) == 3


def test_spans_sampled_all(tracer, test_spans):
    with tracer.trace("root") as span:
        span.sampled = True
        with tracer.trace("child") as span:
            span.sampled = True
        with tracer.trace("child") as span:
            span.sampled = True

    spans = test_spans.pop()
    assert len(spans) == 3


def test_closing_other_context_spans_single_span(tracer, test_spans):
    """
    Ensure that a span created in one thread can be finished in another without
    breaking the active span management.
    """

    def _target(span):
        assert tracer.current_span() is None
        span.finish()
        assert tracer.current_span() is None

    span = tracer.trace("main thread")
    assert tracer.current_span() is span
    t1 = threading.Thread(target=_target, args=(span,))
    t1.start()
    t1.join()
    assert tracer.current_span() is None

    spans = test_spans.pop()
    assert len(spans) == 1


def test_closing_other_context_spans_multi_spans(tracer, test_spans):
    """
    Ensure that spans created in one thread can be finished in another without
    breaking the active span management.
    """

    def _target(span):
        assert tracer.current_span() is None
        span.finish()
        assert tracer.current_span() is None

    root = tracer.trace("root span")
    span = tracer.trace("child span")
    assert tracer.current_span() is span
    t1 = threading.Thread(target=_target, args=(span,))
    t1.start()
    t1.join()
    assert tracer.current_span() is root
    root.finish()

    spans = test_spans.pop()
    assert len(spans) == 2


def test_fork_manual_span_same_context(tracer):
    span = tracer.trace("test")
    pid = os.fork()

    if pid == 0:
        child = tracer.start_span("child", child_of=span)
        assert child.parent_id == span.span_id
        assert child._parent is None
        # No more current span strong reference to avoid memory leaks.
        assert tracer.current_span() is None
        child.finish()
        os._exit(12)

    span.finish()
    _, status = os.waitpid(pid, 0)
    exit_code = os.WEXITSTATUS(status)
    assert exit_code == 12


def test_fork_manual_span_different_contexts(tracer):
    span = tracer.start_span("test")
    pid = os.fork()

    if pid == 0:
        child = tracer.start_span("child", child_of=span)
        assert child.parent_id == span.span_id
        assert child._parent is None
        assert tracer.current_span() is None
        child.finish()
        os._exit(12)

    span.finish()
    _, status = os.waitpid(pid, 0)
    exit_code = os.WEXITSTATUS(status)
    assert exit_code == 12


def test_fork_pid(tracer):
    root = tracer.trace("root_span")
    assert root.get_tag("runtime-id") is not None
    assert root.get_metric(PID) is not None

    pid = os.fork()

    if pid:
        child1 = tracer.trace("span1")
        assert child1.get_tag("runtime-id") is None
        assert child1.get_metric(PID) is None
        child1.finish()
    else:
        child2 = tracer.trace("span2")
        assert child2.get_tag("runtime-id") is not None
        assert child2.get_metric(PID) is not None
        child2.finish()
        os._exit(12)

    root.finish()
    _, status = os.waitpid(pid, 0)
    exit_code = os.WEXITSTATUS(status)
    assert exit_code == 12


def test_tracer_api_version():
    t = Tracer()
    assert isinstance(t._writer._encoder, MsgpackEncoderV03)

    t.configure(api_version="v0.5")
    assert isinstance(t._writer._encoder, MsgpackEncoderV05)

    t.configure(api_version="v0.4")
    assert isinstance(t._writer._encoder, MsgpackEncoderV03)


@pytest.mark.parametrize("enabled", [True, False])
def test_tracer_memory_leak_span_processors(enabled):
    """
    Test whether the tracer or span processors will hold onto
    span references after the trace is complete.

    This is a regression test for the tracer not calling on_span_finish
    of SpanAggregator when the tracer was disabled and traces leaking.
    """
    spans = weakref.WeakSet()

    # Filter to ensure we don't send the traces to the writer
    class DropAllFilter:
        def process_trace(self, trace):
            return None

    t = Tracer()
    t.enabled = enabled
    t.configure(settings={"FILTERS": [DropAllFilter()]})

    for _ in range(5):
        with t.trace("test") as span:
            spans.add(span)

    # Be sure to dereference the last Span held by the local variable `span`
    span = None

    # Force gc
    gc.collect()
    assert len(spans) == 0


def test_top_level(tracer):
    with tracer.trace("parent", service="my-svc") as parent_span:
        assert _is_top_level(parent_span)
        with tracer.trace("child") as child_span:
            assert not _is_top_level(child_span)
            with tracer.trace("subchild") as subchild_span:
                assert not _is_top_level(subchild_span)
            with tracer.trace("subchild2", service="svc-2") as subchild_span2:
                assert _is_top_level(subchild_span2)

    with tracer.trace("parent", service="my-svc") as parent_span:
        assert _is_top_level(parent_span)
        with tracer.trace("child", service="child-svc") as child_span:
            assert _is_top_level(child_span)
        with tracer.trace("child2", service="child-svc") as child_span2:
            assert _is_top_level(child_span2)


def test_ctx_api():
    from ddtrace.internal import _context

    tracer = Tracer()

    with pytest.raises(ValueError):
        _context.get_item("key")
    with pytest.raises(ValueError):
        _context.set_item("key", "val")

    with tracer.trace("root"):
        v = _context.get_item("my.val")
        assert v is None

        _context.set_item("appsec.key", "val")
        _context.set_items({"appsec.key2": "val2", "appsec.key3": "val3"})
        assert _context.get_item("appsec.key") == "val"
        assert _context.get_item("appsec.key2") == "val2"
        assert _context.get_item("appsec.key3") == "val3"
        assert _context.get_items(["appsec.key"]) == ["val"]
        assert _context.get_items(["appsec.key", "appsec.key2", "appsec.key3"]) == ["val", "val2", "val3"]

        with tracer.trace("child"):
            assert _context.get_item("appsec.key") == "val"

    with pytest.raises(ValueError):
        _context.get_item("appsec.key")
    with pytest.raises(ValueError):
        _context.get_items(["appsec.key"])



import sys

import mock
from mock.mock import call
import pytest
from six import PY2

from ddtrace.debugging._function.discovery import FunctionDiscovery
from ddtrace.debugging._function.store import FunctionStore
from ddtrace.internal.utils.inspection import linenos
import tests.submod.stuff as stuff


def gen_wrapper(*arg):
    def _wrapper(wrapped, args, kwargs):
        # take a snapshot of args and kwargs before they are modified by the
        # wrapped function
        try:
            result = wrapped(*args, **kwargs)
            # capture the return value
            mock, v = arg
            mock(v)
            return result
        except Exception:
            # capture the exception
            raise
        finally:
            # finalise and push the snapshot
            pass

    return _wrapper


def test_function_inject():
    with FunctionStore() as store:
        lo = min(linenos(stuff.modulestuff))
        function = FunctionDiscovery.from_module(stuff).at_line(lo)[0]
        hook = mock.Mock()()

        store.inject_hook(function, hook, lo, 42)
        stuff.modulestuff(None)
        hook.assert_called_once_with(42)

    stuff.modulestuff(None)
    hook.assert_called_once_with(42)


def test_function_wrap():
    with FunctionStore() as store:
        function = FunctionDiscovery.from_module(stuff).by_name(stuff.modulestuff.__name__)

        assert function is stuff.modulestuff

        arg = mock.Mock()
        store.wrap(function, gen_wrapper(arg, 42))

        stuff.modulestuff(None)

        arg.assert_called_once_with(42)

    stuff.modulestuff(None)
    arg.assert_called_once_with(42)


def test_function_inject_wrap():
    with FunctionStore() as store:
        # Function retrieval
        lo = min(linenos(stuff.modulestuff))
        function = FunctionDiscovery.from_module(stuff).at_line(lo)[0]
        assert function is FunctionDiscovery.from_module(stuff).by_name(function.__name__)
        assert function is stuff.modulestuff

        # Injection
        hook = mock.Mock()()
        store.inject_hook(function, hook, lo, 42)
        stuff.modulestuff(None)
        hook.assert_called_once_with(42)

        # Wrapping
        arg = mock.Mock()
        store.wrap(stuff.modulestuff, gen_wrapper(arg, 42))
        stuff.modulestuff(None)
        arg.assert_called_once_with(42)

        # Restore
        store.restore_all()
        stuff.modulestuff(None)
        hook.assert_has_calls([call(42), call(42)])
        arg.assert_called_once_with(42)


def test_function_wrap_inject():
    with FunctionStore() as store:
        # Function retrieval
        lo = min(linenos(stuff.modulestuff))
        function = FunctionDiscovery.from_module(stuff).by_name(stuff.modulestuff.__name__)
        assert function is FunctionDiscovery.from_module(stuff).at_line(lo)[0]

        # Wrapping
        arg = mock.Mock()
        store.wrap(function, gen_wrapper(arg, 42))
        stuff.modulestuff(None)
        arg.assert_called_once_with(42)

        # Injection
        hook = mock.Mock()()
        store.inject_hook(function, hook, lo, 42)
        stuff.modulestuff(None)
        hook.assert_called_once_with(42)

        # Restore
        store.restore_all()
        stuff.modulestuff(None)
        hook.assert_called_once_with(42)
        arg.assert_has_calls([call(42), call(42)])


def test_function_wrap_property():
    with FunctionStore() as store:
        function = FunctionDiscovery.from_module(stuff).by_name(
            ".".join((stuff.Stuff.__name__, stuff.Stuff.propertystuff.fget.__name__, "getter"))
        )

        assert function is stuff.Stuff.propertystuff.fget

        arg = mock.Mock()
        store.wrap(function, gen_wrapper(arg, 42))

        s = stuff.Stuff()
        s.propertystuff

        arg.assert_called_once_with(42)

        store.restore_all()
        s.propertystuff
        arg.assert_called_once_with(42)


@pytest.mark.xfail(
    sys.version_info < (3, 6),
    reason="Support for decorated methods is currently not available for this version of Python",
)
def test_function_wrap_decorated():
    with FunctionStore() as store:
        function = FunctionDiscovery.from_module(stuff).by_name(
            ".".join((stuff.Stuff.__name__, stuff.Stuff.doublydecoratedstuff.__name__))
        )

        if PY2:
            assert function.__code__ is stuff.Stuff.doublydecoratedstuff.__func__.__code__
        else:
            assert function is stuff.Stuff.doublydecoratedstuff

        arg = mock.Mock()
        store.wrap(function, gen_wrapper(arg, 42))

        s = stuff.Stuff()
        s.doublydecoratedstuff()

        arg.assert_called_once_with(42)

        store.restore_all()
        s.doublydecoratedstuff()
        arg.assert_called_once_with(42)


def test_function_unwrap():
    with FunctionStore() as store:
        function = FunctionDiscovery.from_module(stuff).by_name(stuff.modulestuff.__name__)
        assert function is stuff.modulestuff
        code = function.__code__

        store.wrap(function, gen_wrapper(None, 42))
        assert code is not stuff.modulestuff.__code__

        store.unwrap(stuff.modulestuff)
        assert code is stuff.modulestuff.__code__


def test_function_inject_wrap_commutativity():
    with FunctionStore() as store:
        code = stuff.modulestuff.__code__

        # Injection
        lo = min(linenos(stuff.modulestuff))
        function = FunctionDiscovery.from_module(stuff).at_line(lo)[0]
        hook = mock.Mock()()
        store.inject_hook(function, hook, lo, 42)

        # Wrapping
        function = FunctionDiscovery.from_module(stuff).by_name(stuff.modulestuff.__name__)
        assert function.__code__ is stuff.modulestuff.__code__
        store.wrap(function, gen_wrapper(None, 42))
        assert code is not stuff.modulestuff.__code__

        # Ejection
        assert stuff.modulestuff.__code__ is not code
        store.eject_hook(stuff.modulestuff, hook, lo, 42)

        # Unwrapping
        store.unwrap(stuff.modulestuff)

        assert stuff.modulestuff.__code__ is not code and stuff.modulestuff.__code__ == code


def test_function_wrap_inject_commutativity():
    with FunctionStore() as store:
        # Wrapping
        function = FunctionDiscovery.from_module(stuff).by_name(stuff.modulestuff.__name__)
        assert function is stuff.modulestuff
        code = function.__code__
        store.wrap(function, gen_wrapper(None, 42))
        assert code is not stuff.modulestuff.__code__

        # Injection
        lo = min(linenos(stuff.modulestuff.__dd_wrapped__))
        function = FunctionDiscovery.from_module(stuff).at_line(lo)[0]
        hook = mock.Mock()()
        store.inject_hook(function, hook, lo, 42)

        # Unwrapping
        store.unwrap(stuff.modulestuff)

        # Ejection
        assert stuff.modulestuff.__code__ is not code
        store.eject_hook(stuff.modulestuff, hook, lo, 42)

        assert stuff.modulestuff.__code__ is not code and stuff.modulestuff.__code__ == code



from functools import partial
import random
import string

from ddtrace import Span
from ddtrace import __version__ as ddtrace_version
from ddtrace.internal.encoding import MSGPACK_ENCODERS


_Span = Span

# DEV: 1.x dropped tracer positional argument
if ddtrace_version.split(".")[0] == "0":
    _Span = partial(_Span, None)

try:
    # the introduction of the buffered encoder changed the internal api
    # see https://github.com/DataDog/dd-trace-py/pull/2422
    from ddtrace.internal._encoding import BufferedEncoder  # noqa: F401

    def init_encoder(encoding, max_size=8 << 20, max_item_size=8 << 20):
        return MSGPACK_ENCODERS[encoding](max_size, max_item_size)


except ImportError:

    def init_encoder(encoding):
        return MSGPACK_ENCODERS[encoding]()


def _rands(size=6, chars=string.ascii_uppercase + string.digits):
    return "".join(random.choice(chars) for _ in range(size))


def _random_values(k, size):
    return list(dict.fromkeys([_rands(size=size) for _ in range(k)]))


def gen_traces(config):
    random.seed(1)

    traces = []

    # choose from a set of randomly generated span attributes
    span_names = _random_values(256, 16)
    resources = _random_values(256, 16)
    services = _random_values(16, 16)
    tag_keys = _random_values(config.ntags, 16)
    metric_keys = _random_values(config.nmetrics, 16)
    dd_origin_values = ["synthetics", "ciapp-test"]

    for _ in range(config.ntraces):
        trace = []
        for i in range(0, config.nspans):
            # first span is root so has no parent otherwise parent is root span
            parent_id = trace[0].span_id if i > 0 else None
            span_name = random.choice(span_names)
            resource = random.choice(resources)
            service = random.choice(services)
            with _Span(span_name, resource=resource, service=service, parent_id=parent_id) as span:
                if i == 0 and config.dd_origin:
                    # Since we're not using the tracer API, a span's context isn't automatically propagated
                    # to its children. The encoder only checks the root span's context in a trace for dd_origin, so
                    # here we need to add dd_origin to the root span's context.
                    span.context.dd_origin = random.choice(dd_origin_values)
                if config.ntags > 0:
                    span.set_tags(dict(zip(tag_keys, [_rands(size=config.ltags) for _ in range(config.ntags)])))
                if config.nmetrics > 0:
                    span.set_metrics(
                        dict(
                            zip(
                                metric_keys,
                                [random.randint(0, 2 ** 16) for _ in range(config.nmetrics)],
                            )
                        )
                    )
                trace.append(span)
        traces.append(trace)
    return traces



from datetime import datetime
from types import TracebackType
from typing import Optional
from typing import Type

from ddtrace.internal import compat


def parse_isoformat(date):
    if date.endswith("Z"):
        try:
            return datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            return datetime.strptime(date, "%Y-%m-%dT%H:%M:%SZ")
    elif hasattr(datetime, "fromisoformat"):
        return datetime.fromisoformat(date)
    raise ValueError("unsupported isoformat")


class StopWatch(object):
    """A simple timer/stopwatch helper class.

    Not thread-safe (when a single watch is mutated by multiple threads at
    the same time). Thread-safe when used by a single thread (not shared) or
    when operations are performed in a thread-safe manner on these objects by
    wrapping those operations with locks.

    It will use the `monotonic`_ pypi library to find an appropriate
    monotonically increasing time providing function (which typically varies
    depending on operating system and Python version).

    .. _monotonic: https://pypi.python.org/pypi/monotonic/
    """

    def __init__(self):
        # type: () -> None
        self._started_at = None  # type: Optional[float]
        self._stopped_at = None  # type: Optional[float]

    def start(self):
        # type: () -> StopWatch
        """Starts the watch."""
        self._started_at = compat.monotonic()
        return self

    def elapsed(self):
        # type: () -> float
        """Get how many seconds have elapsed.

        :return: Number of seconds elapsed
        :rtype: float
        """
        # NOTE: datetime.timedelta does not support nanoseconds, so keep a float here
        if self._started_at is None:
            raise RuntimeError("Can not get the elapsed time of a stopwatch" " if it has not been started/stopped")
        if self._stopped_at is None:
            now = compat.monotonic()
        else:
            now = self._stopped_at
        return now - self._started_at

    def __enter__(self):
        # type: () -> StopWatch
        """Starts the watch."""
        self.start()
        return self

    def __exit__(self, tp, value, traceback):
        # type: (Optional[Type[BaseException]], Optional[BaseException], Optional[TracebackType]) -> None
        """Stops the watch."""
        self.stop()

    def stop(self):
        # type: () -> StopWatch
        """Stops the watch."""
        if self._started_at is None:
            raise RuntimeError("Can not stop a stopwatch that has not been" " started")
        self._stopped_at = compat.monotonic()
        return self



import mock
import pytest

from ddtrace.internal.compat import PY2
from ddtrace.internal.runtime.container import CGroupInfo
from ddtrace.internal.runtime.container import get_container_info

from .utils import cgroup_line_valid_test_cases


# Map expected Py2 exception to Py3 name
if PY2:
    FileNotFoundError = IOError  # noqa: A001


def get_mock_open(read_data=None):
    mock_open = mock.mock_open(read_data=read_data)
    return mock.patch("ddtrace.internal.runtime.container.open", mock_open)


def test_cgroup_info_init():
    # Assert default all attributes to `None`
    info = CGroupInfo()
    for attr in ("id", "groups", "path", "container_id", "controllers", "pod_id"):
        assert getattr(info, attr) is None

    # Assert init with property sets property
    info = CGroupInfo(container_id="test-container-id")
    assert info.container_id == "test-container-id"


@pytest.mark.parametrize(
    "line,expected_info",
    # Valid generated cases + one off cases
    cgroup_line_valid_test_cases()
    + [
        # Valid, extra spaces
        (
            "    13:name=systemd:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860    ",
            CGroupInfo(
                id="13",
                groups="name=systemd",
                controllers=["name=systemd"],
                path="/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860",
                container_id="3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860",
                pod_id=None,
            ),
        ),
        # Valid, bookended newlines
        (
            "\r\n13:name=systemd:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860\r\n",
            CGroupInfo(
                id="13",
                groups="name=systemd",
                controllers=["name=systemd"],
                path="/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860",
                container_id="3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860",
                pod_id=None,
            ),
        ),
        # Valid, fargate >= 1.4.0
        (
            "1:name=systemd:/ecs/34dc0b5e626f2c5c4c5170e34b10e765-1234567890",
            CGroupInfo(
                id="1",
                groups="name=systemd",
                controllers=["name=systemd"],
                path="/ecs/34dc0b5e626f2c5c4c5170e34b10e765-1234567890",
                container_id="34dc0b5e626f2c5c4c5170e34b10e765-1234567890",
                pod_id=None,
            ),
        ),
        # Invalid container_ids
        (
            # One character too short
            "13:name=systemd:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f86986",
            CGroupInfo(
                id="13",
                groups="name=systemd",
                controllers=["name=systemd"],
                path="/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f86986",
                container_id=None,
                pod_id=None,
            ),
        ),
        (
            # Non-hex
            "13:name=systemd:/docker/3726184226f5d3147c25fzyxw5b60097e378e8a720503a5e19ecfdf29f869860",
            CGroupInfo(
                id="13",
                groups="name=systemd",
                controllers=["name=systemd"],
                path="/docker/3726184226f5d3147c25fzyxw5b60097e378e8a720503a5e19ecfdf29f869860",
                container_id=None,
                pod_id=None,
            ),
        ),
        # Invalid id
        (
            # non-digit
            "a:name=systemd:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860",
            None,
        ),
        (
            # missing
            ":name=systemd:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860",
            None,
        ),
        # Missing group
        (
            # empty
            "13::/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860",
            CGroupInfo(
                id="13",
                groups="",
                controllers=[],
                path="/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860",
                container_id="3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860",
                pod_id=None,
            ),
        ),
        (
            # missing
            "13:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860",
            None,
        ),
        # Empty line
        (
            "",
            None,
        ),
    ],
)
def test_cgroup_info_from_line(line, expected_info):
    info = CGroupInfo.from_line(line)
    info2 = CGroupInfo.from_line(line)

    # Check __eq__
    assert info == info2

    if expected_info is None:
        assert info is None, line
    else:
        for attr in ("id", "groups", "path", "container_id", "controllers", "pod_id"):
            assert getattr(info, attr) == getattr(expected_info, attr), line


@pytest.mark.parametrize(
    "file_contents,container_id",
    (
        # Docker file
        (
            """
13:name=systemd:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860
12:pids:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860
11:hugetlb:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860
10:net_prio:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860
9:perf_event:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860
8:net_cls:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860
7:freezer:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860
6:devices:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860
5:memory:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860
4:blkio:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860
3:cpuacct:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860
2:cpu:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860
1:cpuset:/docker/3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860
            """,
            "3726184226f5d3147c25fdeab5b60097e378e8a720503a5e19ecfdf29f869860",
        ),
        # k8s file
        (
            """
11:perf_event:/kubepods/test/pod3d274242-8ee0-11e9-a8a6-1e68d864ef1a/3e74d3fd9db4c9dd921ae05c2502fb984d0cde1b36e581b13f79c639da4518a1
10:pids:/kubepods/test/pod3d274242-8ee0-11e9-a8a6-1e68d864ef1a/3e74d3fd9db4c9dd921ae05c2502fb984d0cde1b36e581b13f79c639da4518a1
9:memory:/kubepods/test/pod3d274242-8ee0-11e9-a8a6-1e68d864ef1a/3e74d3fd9db4c9dd921ae05c2502fb984d0cde1b36e581b13f79c639da4518a1
8:cpu,cpuacct:/kubepods/test/pod3d274242-8ee0-11e9-a8a6-1e68d864ef1a/3e74d3fd9db4c9dd921ae05c2502fb984d0cde1b36e581b13f79c639da4518a1
7:blkio:/kubepods/test/pod3d274242-8ee0-11e9-a8a6-1e68d864ef1a/3e74d3fd9db4c9dd921ae05c2502fb984d0cde1b36e581b13f79c639da4518a1
6:cpuset:/kubepods/test/pod3d274242-8ee0-11e9-a8a6-1e68d864ef1a/3e74d3fd9db4c9dd921ae05c2502fb984d0cde1b36e581b13f79c639da4518a1
5:devices:/kubepods/test/pod3d274242-8ee0-11e9-a8a6-1e68d864ef1a/3e74d3fd9db4c9dd921ae05c2502fb984d0cde1b36e581b13f79c639da4518a1
4:freezer:/kubepods/test/pod3d274242-8ee0-11e9-a8a6-1e68d864ef1a/3e74d3fd9db4c9dd921ae05c2502fb984d0cde1b36e581b13f79c639da4518a1
3:net_cls,net_prio:/kubepods/test/pod3d274242-8ee0-11e9-a8a6-1e68d864ef1a/3e74d3fd9db4c9dd921ae05c2502fb984d0cde1b36e581b13f79c639da4518a1
2:hugetlb:/kubepods/test/pod3d274242-8ee0-11e9-a8a6-1e68d864ef1a/3e74d3fd9db4c9dd921ae05c2502fb984d0cde1b36e581b13f79c639da4518a1
1:name=systemd:/kubepods/test/pod3d274242-8ee0-11e9-a8a6-1e68d864ef1a/3e74d3fd9db4c9dd921ae05c2502fb984d0cde1b36e581b13f79c639da4518a1
            """,
            "3e74d3fd9db4c9dd921ae05c2502fb984d0cde1b36e581b13f79c639da4518a1",
        ),
        # k8 format with additional characters before task ID
        (
            """
1:name=systemd:/kubepods.slice/kubepods-burstable.slice/kubepods-burstable-pod2d3da189_6407_48e3_9ab6_78188d75e609.slice/docker-7b8952daecf4c0e44bbcefe1b5c5ebc7b4839d4eefeccefe694709d3809b6199.scope
            """,
            "7b8952daecf4c0e44bbcefe1b5c5ebc7b4839d4eefeccefe694709d3809b6199",
        ),
        # ECS file
        (
            """
9:perf_event:/ecs/test-ecs-classic/5a0d5ceddf6c44c1928d367a815d890f/38fac3e99302b3622be089dd41e7ccf38aff368a86cc339972075136ee2710ce
8:memory:/ecs/test-ecs-classic/5a0d5ceddf6c44c1928d367a815d890f/38fac3e99302b3622be089dd41e7ccf38aff368a86cc339972075136ee2710ce
7:hugetlb:/ecs/test-ecs-classic/5a0d5ceddf6c44c1928d367a815d890f/38fac3e99302b3622be089dd41e7ccf38aff368a86cc339972075136ee2710ce
6:freezer:/ecs/test-ecs-classic/5a0d5ceddf6c44c1928d367a815d890f/38fac3e99302b3622be089dd41e7ccf38aff368a86cc339972075136ee2710ce
5:devices:/ecs/test-ecs-classic/5a0d5ceddf6c44c1928d367a815d890f/38fac3e99302b3622be089dd41e7ccf38aff368a86cc339972075136ee2710ce
4:cpuset:/ecs/test-ecs-classic/5a0d5ceddf6c44c1928d367a815d890f/38fac3e99302b3622be089dd41e7ccf38aff368a86cc339972075136ee2710ce
3:cpuacct:/ecs/test-ecs-classic/5a0d5ceddf6c44c1928d367a815d890f/38fac3e99302b3622be089dd41e7ccf38aff368a86cc339972075136ee2710ce
2:cpu:/ecs/test-ecs-classic/5a0d5ceddf6c44c1928d367a815d890f/38fac3e99302b3622be089dd41e7ccf38aff368a86cc339972075136ee2710ce
1:blkio:/ecs/test-ecs-classic/5a0d5ceddf6c44c1928d367a815d890f/38fac3e99302b3622be089dd41e7ccf38aff368a86cc339972075136ee2710ce
            """,
            "38fac3e99302b3622be089dd41e7ccf38aff368a86cc339972075136ee2710ce",
        ),
        # Fargate file < 1.4.0
        (
            """
11:hugetlb:/ecs/55091c13-b8cf-4801-b527-f4601742204d/432624d2150b349fe35ba397284dea788c2bf66b885d14dfc1569b01890ca7da
10:pids:/ecs/55091c13-b8cf-4801-b527-f4601742204d/432624d2150b349fe35ba397284dea788c2bf66b885d14dfc1569b01890ca7da
9:cpuset:/ecs/55091c13-b8cf-4801-b527-f4601742204d/432624d2150b349fe35ba397284dea788c2bf66b885d14dfc1569b01890ca7da
8:net_cls,net_prio:/ecs/55091c13-b8cf-4801-b527-f4601742204d/432624d2150b349fe35ba397284dea788c2bf66b885d14dfc1569b01890ca7da
7:cpu,cpuacct:/ecs/55091c13-b8cf-4801-b527-f4601742204d/432624d2150b349fe35ba397284dea788c2bf66b885d14dfc1569b01890ca7da
6:perf_event:/ecs/55091c13-b8cf-4801-b527-f4601742204d/432624d2150b349fe35ba397284dea788c2bf66b885d14dfc1569b01890ca7da
5:freezer:/ecs/55091c13-b8cf-4801-b527-f4601742204d/432624d2150b349fe35ba397284dea788c2bf66b885d14dfc1569b01890ca7da
4:devices:/ecs/55091c13-b8cf-4801-b527-f4601742204d/432624d2150b349fe35ba397284dea788c2bf66b885d14dfc1569b01890ca7da
3:blkio:/ecs/55091c13-b8cf-4801-b527-f4601742204d/432624d2150b349fe35ba397284dea788c2bf66b885d14dfc1569b01890ca7da
2:memory:/ecs/55091c13-b8cf-4801-b527-f4601742204d/432624d2150b349fe35ba397284dea788c2bf66b885d14dfc1569b01890ca7da
1:name=systemd:/ecs/55091c13-b8cf-4801-b527-f4601742204d/432624d2150b349fe35ba397284dea788c2bf66b885d14dfc1569b01890ca7da
            """,
            "432624d2150b349fe35ba397284dea788c2bf66b885d14dfc1569b01890ca7da",
        ),
        # Fargate file >= 1.4.0
        (
            """
11:hugetlb:/ecs/55091c13-b8cf-4801-b527-f4601742204d/34dc0b5e626f2c5c4c5170e34b10e765-1234567890
10:pids:/ecs/55091c13-b8cf-4801-b527-f4601742204d/34dc0b5e626f2c5c4c5170e34b10e765-1234567890
9:cpuset:/ecs/55091c13-b8cf-4801-b527-f4601742204d/34dc0b5e626f2c5c4c5170e34b10e765-1234567890
8:net_cls,net_prio:/ecs/55091c13-b8cf-4801-b527-f4601742204d/34dc0b5e626f2c5c4c5170e34b10e765-1234567890
7:cpu,cpuacct:/ecs/55091c13-b8cf-4801-b527-f4601742204d/34dc0b5e626f2c5c4c5170e34b10e765-1234567890
6:perf_event:/ecs/55091c13-b8cf-4801-b527-f4601742204d/34dc0b5e626f2c5c4c5170e34b10e765-1234567890
5:freezer:/ecs/55091c13-b8cf-4801-b527-f4601742204d/34dc0b5e626f2c5c4c5170e34b10e765-1234567890
4:devices:/ecs/55091c13-b8cf-4801-b527-f4601742204d/34dc0b5e626f2c5c4c5170e34b10e765-1234567890
3:blkio:/ecs/55091c13-b8cf-4801-b527-f4601742204d/34dc0b5e626f2c5c4c5170e34b10e765-1234567890
2:memory:/ecs/55091c13-b8cf-4801-b527-f4601742204d/34dc0b5e626f2c5c4c5170e34b10e765-1234567890
1:name=systemd:/ecs/34dc0b5e626f2c5c4c5170e34b10e765-1234567890
            """,
            "34dc0b5e626f2c5c4c5170e34b10e765-1234567890",
        ),
        # Linux non-containerized file
        (
            """
11:blkio:/user.slice/user-0.slice/session-14.scope
10:memory:/user.slice/user-0.slice/session-14.scope
9:hugetlb:/
8:cpuset:/
7:pids:/user.slice/user-0.slice/session-14.scope
6:freezer:/
5:net_cls,net_prio:/
4:perf_event:/
3:cpu,cpuacct:/user.slice/user-0.slice/session-14.scope
2:devices:/user.slice/user-0.slice/session-14.scope
1:name=systemd:/user.slice/user-0.slice/session-14.scope
            """,
            None,
        ),
        # Empty file
        (
            "",
            None,
        ),
        # Missing file
        (
            None,
            None,
        ),
    ),
)
def test_get_container_info(file_contents, container_id):
    with get_mock_open(read_data=file_contents) as mock_open:
        # simulate the file not being found
        if file_contents is None:
            mock_open.side_effect = FileNotFoundError

        info = get_container_info()

        if container_id is None:
            assert info is None
        else:
            assert info.container_id == container_id

        mock_open.assert_called_once_with("/proc/self/cgroup", mode="r")


@pytest.mark.parametrize(
    "pid,file_name",
    (
        ("13", "/proc/13/cgroup"),
        (13, "/proc/13/cgroup"),
        ("self", "/proc/self/cgroup"),
    ),
)
def test_get_container_info_with_pid(pid, file_name):
    # DEV: We need at least 1 line for the loop to call `CGroupInfo.from_line`
    with get_mock_open(read_data="\r\n") as mock_open:
        assert get_container_info(pid=pid) is None

        mock_open.assert_called_once_with(file_name, mode="r")


@mock.patch("ddtrace.internal.runtime.container.CGroupInfo.from_line")
@mock.patch("ddtrace.internal.runtime.container.log")
def test_get_container_info_exception(mock_log, mock_from_line):
    exception = Exception()
    mock_from_line.side_effect = exception

    # DEV: We need at least 1 line for the loop to call `CGroupInfo.from_line`
    with get_mock_open(read_data="\r\n") as mock_open:
        # Assert calling `get_container_info()` does not bubble up the exception
        assert get_container_info() is None

        # Assert we called everything we expected
        mock_from_line.assert_called_once_with("\r\n")
        mock_open.assert_called_once_with("/proc/self/cgroup", mode="r")

        # Ensure we logged the exception
        mock_log.debug.assert_called_once_with("Failed to parse cgroup file for pid %r", "self", exc_info=True)



from typing import Any


class AttrDict(dict):
    """
    dict implementation that allows for item attribute access


    Example::

       data = AttrDict()
       data['key'] = 'value'
       print(data['key'])

       data.key = 'new-value'
       print(data.key)

       # Convert an existing `dict`
       data = AttrDict(dict(key='value'))
       print(data.key)
    """

    def __getattr__(self, key):
        # type: (str) -> Any
        if key in self:
            return self[key]
        return object.__getattribute__(self, key)

    def __setattr__(self, key, value):
        # type: (str, Any) -> None
        # 1) Ensure if the key exists from a dict key we always prefer that
        # 2) If we do not have an existing key but we do have an attr, set that
        # 3) No existing key or attr exists, so set a key
        if key in self:
            # Update any existing key
            self[key] = value
        elif hasattr(self, key):
            # Allow overwriting an existing attribute, e.g. `self.global_config = dict()`
            object.__setattr__(self, key, value)
        else:
            # Set a new key
            self[key] = value



from unittest import TestCase

import pytest

import ddtrace
from ddtrace.ext import ci
from ddtrace.ext.http import URL
from ddtrace.filters import FilterRequestsOnUrl
from ddtrace.filters import TraceCiVisibilityFilter
from ddtrace.filters import TraceFilter
from ddtrace.span import Span


class TraceCiVisibilityFilterTests(TestCase):
    def test_filters_non_test_root_spans(self):
        trace_filter = TraceCiVisibilityFilter()
        root_test_span = Span(name="span1", span_type="test")
        root_test_span._local_root = root_test_span
        # Root span in trace is a test
        trace = [root_test_span]
        self.assertEqual(trace_filter.process_trace(trace), trace)
        assert root_test_span.get_tag(ci.LIBRARY_VERSION) == ddtrace.__version__

        root_span = Span(name="span1")
        root_span._local_root = root_span
        # Root span in trace is not a test
        trace = [root_span]
        self.assertEqual(trace_filter.process_trace(trace), None)


class FilterRequestOnUrlTests(TestCase):
    def test_is_match(self):
        span = Span(name="Name")
        span.set_tag(URL, r"http://example.com")
        filtr = FilterRequestsOnUrl("http://examp.*.com")
        trace = filtr.process_trace([span])
        self.assertIsNone(trace)

    def test_is_not_match(self):
        span = Span(name="Name")
        span.set_tag(URL, r"http://anotherexample.com")
        filtr = FilterRequestsOnUrl("http://examp.*.com")
        trace = filtr.process_trace([span])
        self.assertIsNotNone(trace)

    def test_list_match(self):
        span = Span(name="Name")
        span.set_tag(URL, r"http://anotherdomain.example.com")
        filtr = FilterRequestsOnUrl([r"http://domain\.example\.com", r"http://anotherdomain\.example\.com"])
        trace = filtr.process_trace([span])
        self.assertIsNone(trace)

    def test_list_no_match(self):
        span = Span(name="Name")
        span.set_tag(URL, r"http://cooldomain.example.com")
        filtr = FilterRequestsOnUrl([r"http://domain\.example\.com", r"http://anotherdomain\.example\.com"])
        trace = filtr.process_trace([span])
        self.assertIsNotNone(trace)


def test_not_implemented_trace_filter():
    class Filter(TraceFilter):
        pass

    with pytest.raises(TypeError):
        Filter()



# 3p
import asyncio

import aiopg.connection
import psycopg2.extensions

from ddtrace.contrib.aiopg.connection import AIOTracedConnection
from ddtrace.contrib.psycopg.patch import _patch_extensions
from ddtrace.contrib.psycopg.patch import _unpatch_extensions
from ddtrace.contrib.psycopg.patch import patch_conn as psycopg_patch_conn
from ddtrace.internal.utils.wrappers import unwrap as _u
from ddtrace.vendor import wrapt


def patch():
    """Patch monkey patches psycopg's connection function
    so that the connection's functions are traced.
    """
    if getattr(aiopg, "_datadog_patch", False):
        return
    setattr(aiopg, "_datadog_patch", True)

    wrapt.wrap_function_wrapper(aiopg.connection, "_connect", patched_connect)
    _patch_extensions(_aiopg_extensions)  # do this early just in case


def unpatch():
    if getattr(aiopg, "_datadog_patch", False):
        setattr(aiopg, "_datadog_patch", False)
        _u(aiopg.connection, "_connect")
        _unpatch_extensions(_aiopg_extensions)


@asyncio.coroutine
def patched_connect(connect_func, _, args, kwargs):
    conn = yield from connect_func(*args, **kwargs)
    return psycopg_patch_conn(conn, traced_conn_cls=AIOTracedConnection)


def _extensions_register_type(func, _, args, kwargs):
    def _unroll_args(obj, scope=None):
        return obj, scope

    obj, scope = _unroll_args(*args, **kwargs)

    # register_type performs a c-level check of the object
    # type so we must be sure to pass in the actual db connection
    if scope and isinstance(scope, wrapt.ObjectProxy):
        scope = scope.__wrapped__._conn

    return func(obj, scope) if scope else func(obj)


# extension hooks
_aiopg_extensions = [
    (psycopg2.extensions.register_type, psycopg2.extensions, "register_type", _extensions_register_type),
]



import httpretty
import pytest

from ddtrace.internal.utils.http import connector


@pytest.mark.parametrize("scheme", ["http", "https", "unix"])
def test_connector(scheme):
    with httpretty.enabled():
        httpretty.register_uri(httpretty.GET, "%s://localhost:8181/api/test" % scheme, body='{"hello": "world"}')

        connect = connector("%s://localhost:8181" % scheme)
        with connect() as conn:
            conn.request("GET", "/api/test")
            response = conn.getresponse()
            assert response.status == 200
            assert response.read() == b'{"hello": "world"}'



from envier import En

from ddtrace import config
from ddtrace.internal.agent import get_trace_url
from ddtrace.internal.constants import DEFAULT_SERVICE_NAME
from ddtrace.internal.utils.config import get_application_name
from ddtrace.version import get_version


DEFAULT_MAX_PROBES = 100
DEFAULT_GLOBAL_RATE_LIMIT = 100.0


def _derive_tags(c):
    # type: (En) -> str
    _tags = dict(env=config.env, version=config.version, debugger_version=get_version())
    _tags.update(config.tags)

    return ",".join([":".join((k, v)) for (k, v) in _tags.items() if v is not None])


class DynamicInstrumentationConfig(En):
    __prefix__ = "dd.dynamic_instrumentation"

    service_name = En.d(str, lambda _: config.service or get_application_name() or DEFAULT_SERVICE_NAME)
    _snapshot_intake_url = En.d(str, lambda _: get_trace_url())
    max_probes = En.d(int, lambda _: DEFAULT_MAX_PROBES)
    global_rate_limit = En.d(float, lambda _: DEFAULT_GLOBAL_RATE_LIMIT)
    _tags_in_qs = En.d(bool, lambda _: True)
    _snapshot_intake_endpoint = En.d(str, lambda _: "/debugger/v1/input")
    tags = En.d(str, _derive_tags)

    enabled = En.v(
        bool,
        "enabled",
        default=False,
        help_type="Boolean",
        help="Enable the debugger",
    )

    metrics = En.v(
        bool,
        "metrics.enabled",
        default=True,
        help_type="Boolean",
        help="Enable diagnostic metrics",
    )

    max_payload_size = En.v(
        int,
        "max_payload_size",
        default=1 << 20,  # 1 MB
        help_type="Integer",
        help="Maximum size in bytes of a single configuration payload that can be handled per request",
    )

    upload_timeout = En.v(
        int,
        "upload.timeout",
        default=30,  # seconds
        help_type="Integer",
        help="Timeout in seconds for uploading a snapshot",
    )

    upload_flush_interval = En.v(
        float,
        "upload.flush_interval",
        default=1.0,  # seconds
        help_type="Float",
        help="Interval in seconds for flushing the snapshot upload queue.",
    )

    diagnostics_interval = En.v(
        int,
        "diagnostics.interval",
        default=3600,  # 1 hour
        help_type="Integer",
        help="Interval in seconds for periodically sending probe diagnostic messages",
    )



import pytest

from ddtrace.debugging._function.discovery import FunctionDiscovery
from ddtrace.internal.compat import PY2
import tests.submod.stuff as stuff


def _f(f):
    if PY2:
        return f.im_func
    return f


@pytest.fixture
def stuff_discovery():
    return FunctionDiscovery.from_module(stuff)


def test_abs_stuff():
    import tests.submod.absstuff as absstuff

    assert sorted(FunctionDiscovery.from_module(absstuff).keys()) == [9, 13, 18, 21]


def test_function_discovery(stuff_discovery):
    assert len(stuff_discovery.at_line(11)) == 3
    assert not stuff_discovery.at_line(69)
    assert len(stuff_discovery.at_line(71)) == 1
    assert not stuff_discovery.at_line(72)
    assert len(stuff_discovery.at_line(86)) == 1
    assert len(stuff_discovery.at_line(89)) == 1


@pytest.mark.parametrize(
    "alias",
    [
        "modulestuff",
        "__specialstuff__",
        "alias",
        "Stuff.propertystuff.getter",
        "Stuff.instancestuff",
        "Stuff.staticstuff",
        "Stuff.classstuff",
        "Stuff.decoratedstuff",
        "Stuff.doublydecoratedstuff",
        "Stuff.nestedstuff",
        "Stuff._Stuff__mangledstuff",
        "Stuff.__mangledstuff",
        "Stuff.generatorstuff",
    ],
)
def test_function_by_name(stuff_discovery, alias):
    assert stuff_discovery.by_name(alias)


@pytest.mark.parametrize(
    "orig,alias",
    [
        ("modulestuff", "alias"),
        ("AliasStuff.foo", "AliasStuff.bar"),
    ],
)
def test_function_aliasing(stuff_discovery, orig, alias):
    assert stuff_discovery.by_name(orig) == stuff_discovery.by_name(alias)


def test_function_property(stuff_discovery):
    original_property = stuff.Stuff.propertystuff
    assert isinstance(original_property, property)

    (original_fget,) = stuff_discovery.at_line(40)
    assert original_property.fget is original_fget


def test_function_staticmethod(stuff_discovery):
    original_sm = stuff.Stuff.staticstuff
    (original_func,) = stuff_discovery.at_line(29)
    assert original_sm is original_func


def test_function_classmethod(stuff_discovery):
    original_cm = stuff.Stuff.classstuff
    (original_func,) = stuff_discovery.at_line(33)
    assert original_cm.__func__ == original_func


def test_function_module_method(stuff_discovery):
    (original_func,) = stuff_discovery.at_line(6)
    assert stuff.modulestuff is original_func


def test_function_instance_method(stuff_discovery):
    cls = stuff.Stuff
    (original_func,) = stuff_discovery.at_line(36)
    assert _f(cls.instancestuff) is original_func


def test_function_decorated_method(stuff_discovery):
    method = stuff.Stuff.decoratedstuff
    (original_func,) = stuff_discovery.at_line(48)
    assert original_func in {_.cell_contents for _ in method.__closure__}


def test_function_mangled(stuff_discovery):
    original_method = stuff.Stuff._Stuff__mangledstuff
    (original_func,) = stuff_discovery.at_line(75)
    assert _f(original_method) is original_func


def test_discovery_after_external_wrapping(stuff):
    import ddtrace.vendor.wrapt as wrapt

    def wrapper(wrapped, inst, args, kwargs):
        pass

    wrapt.wrap_function_wrapper(stuff, "Stuff.instancestuff", wrapper)
    assert isinstance(stuff.Stuff.instancestuff, (wrapt.BoundFunctionWrapper, wrapt.FunctionWrapper))

    code = stuff.Stuff.instancestuff.__code__
    f = FunctionDiscovery(stuff)[36][0]

    assert isinstance(f, (wrapt.BoundFunctionWrapper, wrapt.FunctionWrapper))
    assert f.__code__ is code


def test_property_non_function_getter(stuff_discovery):
    with pytest.raises(ValueError):
        stuff_discovery.by_name("PropertyStuff.foo")



import json
import logging
import os

from paste import fixture
from paste.deploy import loadapp
import pylons
import pytest
from routes import url_for

from ddtrace import config
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.constants import ERROR_MSG
from ddtrace.constants import ERROR_STACK
from ddtrace.constants import ERROR_TYPE
from ddtrace.constants import SAMPLING_PRIORITY_KEY
from ddtrace.contrib.pylons import PylonsTraceMiddleware
from ddtrace.ext import http
from ddtrace.ext import user
from ddtrace.internal import _context
from ddtrace.internal.compat import urlencode
from tests.appsec.test_processor import RULES_GOOD_PATH
from tests.opentracer.utils import init_tracer
from tests.utils import TracerTestCase
from tests.utils import assert_is_measured
from tests.utils import assert_span_http_status_code
from tests.utils import override_env
from tests.utils import override_global_config


class PylonsTestCase(TracerTestCase):
    """Pylons Test Controller that is used to test specific
    cases defined in the Pylons controller. To test a new behavior,
    add a new action in the `app.controllers.root` module.
    """

    conf_dir = os.path.dirname(os.path.abspath(__file__))

    @pytest.fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    def setUp(self):
        super(PylonsTestCase, self).setUp()
        # initialize a real traced Pylons app
        wsgiapp = loadapp("config:test.ini", relative_to=PylonsTestCase.conf_dir)
        self._wsgiapp = wsgiapp
        app = PylonsTraceMiddleware(wsgiapp, self.tracer, service="web")
        self.app = fixture.TestApp(app)

    def test_controller_exception(self):
        """Ensure exceptions thrown in controllers can be handled.

        No error tags should be set in the span.
        """
        from .app.middleware import ExceptionToSuccessMiddleware

        wsgiapp = ExceptionToSuccessMiddleware(self._wsgiapp)
        app = PylonsTraceMiddleware(wsgiapp, self.tracer, service="web")

        app = fixture.TestApp(app)
        app.get(url_for(controller="root", action="raise_exception"))

        spans = self.pop_spans()

        assert spans, spans
        assert len(spans) == 1
        span = spans[0]

        assert_is_measured(span)
        assert span.service == "web"
        assert span.resource == "root.raise_exception"
        assert span.error == 0
        assert span.get_tag(http.URL) == "http://localhost:80/raise_exception"
        assert_span_http_status_code(span, 200)
        assert http.QUERY_STRING not in span.get_tags()
        assert span.get_tag(ERROR_MSG) is None
        assert span.get_tag(ERROR_TYPE) is None
        assert span.get_tag(ERROR_STACK) is None
        assert span.span_type == "web"

    def test_mw_exc_success(self):
        """Ensure exceptions can be properly handled by other middleware.

        No error should be reported in the span.
        """
        from .app.middleware import ExceptionMiddleware
        from .app.middleware import ExceptionToSuccessMiddleware

        wsgiapp = ExceptionMiddleware(self._wsgiapp)
        wsgiapp = ExceptionToSuccessMiddleware(wsgiapp)
        app = PylonsTraceMiddleware(wsgiapp, self.tracer, service="web")
        app = fixture.TestApp(app)

        app.get(url_for(controller="root", action="index"))

        spans = self.pop_spans()

        assert spans, spans
        assert len(spans) == 1
        span = spans[0]

        assert_is_measured(span)
        assert span.service == "web"
        assert span.resource == "None.None"
        assert span.error == 0
        assert span.get_tag(http.URL) == "http://localhost:80/"
        assert_span_http_status_code(span, 200)
        assert span.get_tag(ERROR_MSG) is None
        assert span.get_tag(ERROR_TYPE) is None
        assert span.get_tag(ERROR_STACK) is None

    def test_middleware_exception(self):
        """Ensure exceptions raised in middleware are properly handled.

        Uncaught exceptions should result in error tagged spans.
        """
        from .app.middleware import ExceptionMiddleware

        wsgiapp = ExceptionMiddleware(self._wsgiapp)
        app = PylonsTraceMiddleware(wsgiapp, self.tracer, service="web")
        app = fixture.TestApp(app)

        with pytest.raises(Exception):
            app.get(url_for(controller="root", action="index"))

        spans = self.pop_spans()

        assert spans, spans
        assert len(spans) == 1
        span = spans[0]

        assert_is_measured(span)
        assert span.service == "web"
        assert span.resource == "None.None"
        assert span.error == 1
        assert span.get_tag(http.URL) == "http://localhost:80/"
        assert_span_http_status_code(span, 500)
        assert span.get_tag(ERROR_MSG) == "Middleware exception"
        assert span.get_tag(ERROR_TYPE) == "exceptions.Exception"
        assert span.get_tag(ERROR_STACK)

    def test_exc_success(self):
        from .app.middleware import ExceptionToSuccessMiddleware

        wsgiapp = ExceptionToSuccessMiddleware(self._wsgiapp)
        app = PylonsTraceMiddleware(wsgiapp, self.tracer, service="web")
        app = fixture.TestApp(app)

        app.get(url_for(controller="root", action="raise_exception"))

        spans = self.pop_spans()
        assert spans, spans
        assert len(spans) == 1
        span = spans[0]

        assert_is_measured(span)
        assert span.service == "web"
        assert span.resource == "root.raise_exception"
        assert span.error == 0
        assert span.get_tag(http.URL) == "http://localhost:80/raise_exception"
        assert_span_http_status_code(span, 200)
        assert span.get_tag(ERROR_MSG) is None
        assert span.get_tag(ERROR_TYPE) is None
        assert span.get_tag(ERROR_STACK) is None

    def test_exc_client_failure(self):
        from .app.middleware import ExceptionToClientErrorMiddleware

        wsgiapp = ExceptionToClientErrorMiddleware(self._wsgiapp)
        app = PylonsTraceMiddleware(wsgiapp, self.tracer, service="web")
        app = fixture.TestApp(app)

        app.get(url_for(controller="root", action="raise_exception"), status=404)

        spans = self.pop_spans()
        assert spans, spans
        assert len(spans) == 1
        span = spans[0]

        assert_is_measured(span)
        assert span.service == "web"
        assert span.resource == "root.raise_exception"
        assert span.error == 0
        assert span.get_tag(http.URL) == "http://localhost:80/raise_exception"
        assert_span_http_status_code(span, 404)
        assert span.get_tag(ERROR_MSG) is None
        assert span.get_tag(ERROR_TYPE) is None
        assert span.get_tag(ERROR_STACK) is None

    def test_success_200(self, query_string=""):
        with override_global_config(dict(_appsec_enabled=True)):
            if query_string:
                fqs = "?" + query_string
            else:
                fqs = ""
            res = self.app.get(url_for(controller="root", action="index") + fqs)
            assert res.status == 200

            spans = self.pop_spans()
            assert spans, spans
            assert len(spans) == 1
            span = spans[0]

            assert_is_measured(span)
            assert span.service == "web"
            assert span.resource == "root.index"
            assert_span_http_status_code(span, 200)
            if config.pylons.trace_query_string:
                assert span.get_tag(http.QUERY_STRING) == query_string
                if config._appsec:
                    assert _context.get_item("http.request.uri", span=span) == "http://localhost:80/?" + query_string
            else:
                assert http.QUERY_STRING not in span.get_tags()
                if config._appsec:
                    assert _context.get_item("http.request.uri", span=span) == "http://localhost:80/"
            assert span.error == 0

    def test_query_string(self):
        return self.test_success_200("foo=bar")

    def test_multi_query_string(self):
        return self.test_success_200("foo=bar&foo=baz&x=y")

    def test_query_string_trace(self):
        with self.override_http_config("pylons", dict(trace_query_string=True)):
            return self.test_success_200("foo=bar")

    def test_multi_query_string_trace(self):
        with self.override_http_config("pylons", dict(trace_query_string=True)):
            return self.test_success_200("foo=bar&foo=baz&x=y")

    def test_appsec_http_raw_uri(self):
        with self.override_global_config(dict(appsec_enabled=True)):
            self.test_query_string()
            self.test_multi_query_string()
            self.test_query_string_trace()
            self.test_multi_query_string_trace()

    def test_analytics_global_on_integration_default(self):
        """
        When making a request
            When an integration trace search is not event sample rate is not set and globally trace search is enabled
                We expect the root span to have the appropriate tag
        """
        with self.override_global_config(dict(analytics_enabled=True)):
            res = self.app.get(url_for(controller="root", action="index"))
            self.assertEqual(res.status, 200)

        self.assert_structure(dict(name="pylons.request", metrics={ANALYTICS_SAMPLE_RATE_KEY: 1.0}))

    def test_analytics_global_on_integration_on(self):
        """
        When making a request
            When an integration trace search is enabled and sample rate is set and globally trace search is enabled
                We expect the root span to have the appropriate tag
        """
        with self.override_global_config(dict(analytics_enabled=True)):
            with self.override_config("pylons", dict(analytics_enabled=True, analytics_sample_rate=0.5)):
                res = self.app.get(url_for(controller="root", action="index"))
                self.assertEqual(res.status, 200)

        self.assert_structure(dict(name="pylons.request", metrics={ANALYTICS_SAMPLE_RATE_KEY: 0.5}))

    def test_analytics_global_off_integration_default(self):
        """
        When making a request
            When an integration trace search is not set and sample rate is set and globally trace search is disabled
                We expect the root span to not include tag
        """
        with self.override_global_config(dict(analytics_enabled=False)):
            res = self.app.get(url_for(controller="root", action="index"))
            self.assertEqual(res.status, 200)

        root = self.get_root_span()
        self.assertIsNone(root.get_metric(ANALYTICS_SAMPLE_RATE_KEY))

    def test_analytics_global_off_integration_on(self):
        """
        When making a request
            When an integration trace search is enabled and sample rate is set and globally trace search is disabled
                We expect the root span to have the appropriate tag
        """
        with self.override_global_config(dict(analytics_enabled=False)):
            with self.override_config("pylons", dict(analytics_enabled=True, analytics_sample_rate=0.5)):
                res = self.app.get(url_for(controller="root", action="index"))
                self.assertEqual(res.status, 200)

        self.assert_structure(dict(name="pylons.request", metrics={ANALYTICS_SAMPLE_RATE_KEY: 0.5}))

    def test_template_render(self):
        res = self.app.get(url_for(controller="root", action="render"))
        assert res.status == 200

        spans = self.pop_spans()
        assert spans, spans
        assert len(spans) == 2
        request = spans[0]
        template = spans[1]

        assert request.service == "web"
        assert request.resource == "root.render"
        assert_span_http_status_code(request, 200)
        assert request.error == 0

        assert template.service == "web"
        assert template.resource == "pylons.render"
        assert template.get_tag("template.name") == "/template.mako"
        assert template.error == 0

    def test_template_render_exception(self):
        with pytest.raises(Exception):
            self.app.get(url_for(controller="root", action="render_exception"))

        spans = self.pop_spans()
        assert spans, spans
        assert len(spans) == 2
        request = spans[0]
        template = spans[1]

        assert request.service == "web"
        assert request.resource == "root.render_exception"
        assert_span_http_status_code(request, 500)
        assert request.error == 1

        assert template.service == "web"
        assert template.resource == "pylons.render"
        assert template.get_tag("template.name") == "/exception.mako"
        assert template.error == 1
        assert template.get_tag("error.msg") == "integer division or modulo by zero"
        assert "ZeroDivisionError: integer division or modulo by zero" in template.get_tag("error.stack")

    def test_failure_500(self):
        with pytest.raises(Exception):
            self.app.get(url_for(controller="root", action="raise_exception"))

        spans = self.pop_spans()
        assert spans, spans
        assert len(spans) == 1
        span = spans[0]

        assert span.service == "web"
        assert span.resource == "root.raise_exception"
        assert span.error == 1
        assert_span_http_status_code(span, 500)
        assert span.get_tag("error.msg") == "Ouch!"
        assert span.get_tag(http.URL) == "http://localhost:80/raise_exception"
        assert "Exception: Ouch!" in span.get_tag("error.stack")

    def test_failure_500_with_wrong_code(self):
        with pytest.raises(Exception):
            self.app.get(url_for(controller="root", action="raise_wrong_code"))

        spans = self.pop_spans()
        assert spans, spans
        assert len(spans) == 1
        span = spans[0]

        assert span.service == "web"
        assert span.resource == "root.raise_wrong_code"
        assert span.error == 1
        assert_span_http_status_code(span, 500)
        assert span.get_tag(http.URL) == "http://localhost:80/raise_wrong_code"
        assert span.get_tag("error.msg") == "Ouch!"
        assert "Exception: Ouch!" in span.get_tag("error.stack")

    def test_failure_500_with_custom_code(self):
        with pytest.raises(Exception):
            self.app.get(url_for(controller="root", action="raise_custom_code"))

        spans = self.pop_spans()
        assert spans, spans
        assert len(spans) == 1
        span = spans[0]

        assert span.service == "web"
        assert span.resource == "root.raise_custom_code"
        assert span.error == 1
        assert_span_http_status_code(span, 512)
        assert span.get_tag(http.URL) == "http://localhost:80/raise_custom_code"
        assert span.get_tag("error.msg") == "Ouch!"
        assert "Exception: Ouch!" in span.get_tag("error.stack")

    def test_failure_500_with_code_method(self):
        with pytest.raises(Exception):
            self.app.get(url_for(controller="root", action="raise_code_method"))

        spans = self.pop_spans()
        assert spans, spans
        assert len(spans) == 1
        span = spans[0]

        assert span.service == "web"
        assert span.resource == "root.raise_code_method"
        assert span.error == 1
        assert_span_http_status_code(span, 500)
        assert span.get_tag(http.URL) == "http://localhost:80/raise_code_method"
        assert span.get_tag("error.msg") == "Ouch!"

    def test_distributed_tracing_default(self):
        # ensure by default, distributed tracing is enabled
        headers = {
            "x-datadog-trace-id": "100",
            "x-datadog-parent-id": "42",
            "x-datadog-sampling-priority": "2",
        }
        res = self.app.get(url_for(controller="root", action="index"), headers=headers)
        assert res.status == 200

        spans = self.pop_spans()
        assert spans, spans
        assert len(spans) == 1
        span = spans[0]

        assert span.trace_id == 100
        assert span.parent_id == 42
        assert span.get_metric(SAMPLING_PRIORITY_KEY) == 2

    def test_distributed_tracing_disabled_via_int_config(self):
        config.pylons["distributed_tracing"] = False
        headers = {
            "x-datadog-trace-id": "100",
            "x-datadog-parent-id": "42",
            "x-datadog-sampling-priority": "2",
        }

        res = self.app.get(url_for(controller="root", action="index"), headers=headers)
        assert res.status == 200

        spans = self.pop_spans()
        assert spans, spans
        assert len(spans) == 1
        span = spans[0]

        assert span.trace_id != 100
        assert span.parent_id != 42
        assert span.get_metric(SAMPLING_PRIORITY_KEY) != 2

    @TracerTestCase.run_in_subprocess(env_overrides=dict(DD_PYLONS_DISTRIBUTED_TRACING="False"))
    def test_distributed_tracing_disabled_via_env_var(self):
        headers = {
            "x-datadog-trace-id": "100",
            "x-datadog-parent-id": "42",
            "x-datadog-sampling-priority": "2",
        }

        res = self.app.get(url_for(controller="root", action="index"), headers=headers)
        assert res.status == 200

        spans = self.pop_spans()
        assert spans, spans
        assert len(spans) == 1
        span = spans[0]

        assert span.trace_id != 100
        assert span.parent_id != 42
        assert span.get_metric(SAMPLING_PRIORITY_KEY) != 2

    def test_success_200_ot(self):
        """OpenTracing version of test_success_200."""
        ot_tracer = init_tracer("pylons_svc", self.tracer)

        with ot_tracer.start_active_span("pylons_get"):
            res = self.app.get(url_for(controller="root", action="index"))
            assert res.status == 200

        spans = self.pop_spans()
        assert spans, spans
        assert len(spans) == 2
        ot_span, dd_span = spans

        # confirm the parenting
        assert ot_span.parent_id is None
        assert dd_span.parent_id == ot_span.span_id

        assert ot_span.name == "pylons_get"
        assert ot_span.service == "pylons_svc"

        assert dd_span.service == "web"
        assert dd_span.resource == "root.index"
        assert_span_http_status_code(dd_span, 200)
        assert dd_span.get_tag(http.URL) == "http://localhost:80/"
        assert dd_span.error == 0

    def test_request_headers(self):
        headers = {
            "my-header": "value",
            "user-agent": "Agent/10.10",
        }
        config.pylons.http.trace_headers(["my-header"])
        res = self.app.get(url_for(controller="root", action="index"), headers=headers)
        assert res.status == 200
        spans = self.pop_spans()
        assert spans[0].get_tag("http.request.headers.my-header") == "value"
        assert spans[0].get_tag(http.USER_AGENT) == "Agent/10.10"

    def test_response_headers(self):
        config.pylons.http.trace_headers(["content-length", "custom-header"])
        res = self.app.get(url_for(controller="root", action="response_headers"))
        assert res.status == 200
        spans = self.pop_spans()
        # Pylons 0.96 and below don't report the content length
        if pylons.__version__ > (0, 9, 6):
            assert spans[0].get_tag("http.response.headers.content-length") == "2"
        assert spans[0].get_tag("http.response.headers.custom-header") == "value"

    def test_pylons_cookie_sql_injection(self):
        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)):
            self.tracer._appsec_enabled = True
            # Hack: need to pass an argument to configure so that the processors are recreated
            self.tracer.configure(api_version="v0.4")
            self.app.cookies = {"attack": "w00tw00t.at.isc.sans.dfind"}
            self.app.get(url_for(controller="root", action="index"))

            spans = self.pop_spans()
            root_span = spans[0]

            appsec_json = root_span.get_tag("_dd.appsec.json")
            assert "triggers" in json.loads(appsec_json if appsec_json else "{}")

            span = _context.get_item("http.request.cookies", span=root_span)
            assert span["attack"] == "w00tw00t.at.isc.sans.dfind"

    def test_pylons_cookie(self):
        with override_global_config(dict(_appsec_enabled=True)):
            self.tracer._appsec_enabled = True
            # Hack: need to pass an argument to configure so that the processors are recreated
            self.tracer.configure(api_version="v0.4")
            self.app.cookies = {"testingcookie_key": "testingcookie_value"}
            self.app.get(url_for(controller="root", action="index"))

            spans = self.pop_spans()
            root_span = spans[0]

            assert root_span.get_tag("_dd.appsec.json") is None
            span = _context.get_item("http.request.cookies", span=root_span)
            assert span["testingcookie_key"] == "testingcookie_value"

    def test_pylons_body_urlencoded(self):
        with self.override_global_config(dict(_appsec_enabled=True)):

            self.tracer.configure(api_version="v0.4")
            payload = urlencode({"mytestingbody_key": "mytestingbody_value"})
            response = self.app.post(
                url_for(controller="root", action="body"),
                params=payload,
                extra_environ={"CONTENT_TYPE": "application/x-www-form-urlencoded"},
            )
            assert response.status == 200

            spans = self.pop_spans()
            assert spans

            root_span = spans[0]
            assert root_span
            assert root_span.get_tag("_dd.appsec.json") is None

            span = dict(_context.get_item("http.request.body", span=root_span))
            assert span
            assert span["mytestingbody_key"] == "mytestingbody_value"

    def test_pylons_request_body_urlencoded_appsec_disabled_then_no_body(self):
        self.tracer._appsec_enabled = False
        # Hack: need to pass an argument to configure so that the processors are recreated
        self.tracer.configure(api_version="v0.4")
        payload = urlencode({"mytestingbody_key": "mytestingbody_value"})
        self.app.post(url_for(controller="root", action="index"), params=payload)

        spans = self.pop_spans()
        root_span = spans[0]

        assert root_span
        assert not _context.get_item("http.request.body", span=root_span)

    def test_pylons_body_urlencoded_attack(self):
        with self.override_global_config(dict(_appsec_enabled=True)):
            with override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)):
                self.tracer._appsec_enabled = True
                # Hack: need to pass an argument to configure so that the processors are recreated
                self.tracer.configure(api_version="v0.4")
                payload = urlencode({"attack": "1' or '1' = '1'"})
                self.app.post(url_for(controller="root", action="index"), params=payload)

                spans = self.pop_spans()
                assert spans

                root_span = spans[0]
                assert root_span

                appsec_json = root_span.get_tag("_dd.appsec.json")
                assert appsec_json
                assert "triggers" in json.loads(appsec_json if appsec_json else "{}")

                query = dict(_context.get_item("http.request.body", span=root_span))
                assert query == {"attack": "1' or '1' = '1'"}

    def test_pylons_body_json(self):
        with override_global_config(dict(_appsec_enabled=True)):
            # Hack: need to pass an argument to configure so that the processors are recreated
            self.tracer.configure(api_version="v0.4")
            payload = json.dumps({"mytestingbody_key": "mytestingbody_value"})
            response = self.app.post(
                url_for(controller="root", action="body"),
                params=payload,
                extra_environ={"CONTENT_TYPE": "application/json"},
            )
            assert response.status == 200

            spans = self.pop_spans()
            assert spans

            root_span = spans[0]
            assert root_span
            assert root_span.get_tag("_dd.appsec.json") is None

            span = dict(_context.get_item("http.request.body", span=root_span))
            assert span
            assert span["mytestingbody_key"] == "mytestingbody_value"

    def test_pylons_body_json_attack(self):
        with self.override_global_config(dict(_appsec_enabled=True)):
            with override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)):
                self.tracer._appsec_enabled = True
                # Hack: need to pass an argument to configure so that the processors are recreated
                self.tracer.configure(api_version="v0.4")
                payload = json.dumps({"attack": "1' or '1' = '1'"})
                self.app.post(
                    url_for(controller="root", action="index"),
                    params=payload,
                    extra_environ={"CONTENT_TYPE": "application/json"},
                )

                spans = self.pop_spans()
                assert spans

                root_span = spans[0]
                appsec_json = root_span.get_tag("_dd.appsec.json")
                assert appsec_json
                assert "triggers" in json.loads(appsec_json if appsec_json else "{}")

                span = dict(_context.get_item("http.request.body", span=root_span))
                assert span
                assert span == {"attack": "1' or '1' = '1'"}

    def test_pylons_body_xml(self):
        with override_global_config(dict(_appsec_enabled=True)):
            # Hack: need to pass an argument to configure so that the processors are recreated
            self.tracer.configure(api_version="v0.4")
            payload = "<mytestingbody_key>mytestingbody_value</mytestingbody_key>"

            response = self.app.post(
                url_for(controller="root", action="body"),
                params=payload,
                extra_environ={"CONTENT_TYPE": "application/xml"},
            )
            assert response.status == 200

            spans = self.pop_spans()
            assert spans

            root_span = spans[0]
            assert root_span
            assert root_span.get_tag("_dd.appsec.json") is None

            span = dict(_context.get_item("http.request.body", span=root_span))
            assert span
            assert span["mytestingbody_key"] == "mytestingbody_value"

    def test_pylons_body_xml_attack(self):
        with override_global_config(dict(_appsec_enabled=True)):
            # Hack: need to pass an argument to configure so that the processors are recreated
            self.tracer.configure(api_version="v0.4")
            payload = "<attack>1' or '1' = '1'</attack>"
            self.app.post(
                url_for(controller="root", action="index"),
                params=payload,
                extra_environ={"CONTENT_TYPE": "application/xml"},
            )

            spans = self.pop_spans()
            assert spans

            root_span = spans[0]
            assert root_span
            assert root_span.get_tag("_dd.appsec.json") is None

            span = dict(_context.get_item("http.request.body", span=root_span))
            assert span
            assert span == {"attack": "1' or '1' = '1'"}

    def test_pylons_body_plain(self):
        with self.override_global_config(dict(_appsec_enabled=True)):
            self.tracer._appsec_enabled = True
            # Hack: need to pass an argument to configure so that the processors are recreated
            self.tracer.configure(api_version="v0.4")
            payload = "foo=bar"

            response = self.app.post(
                url_for(controller="root", action="body"), params=payload, extra_environ={"CONTENT_TYPE": "text/plain"}
            )
            assert response.status == 200

            spans = self.pop_spans()
            assert spans

            root_span = spans[0]
            assert root_span
            assert root_span.get_tag("_dd.appsec.json") is None

            span = _context.get_item("http.request.body", span=root_span)
            assert span
            assert span == "foo=bar"

    def test_pylons_body_plain_attack(self):
        with self.override_global_config(dict(_appsec_enabled=True)):
            self.tracer._appsec_enabled = True
            # Hack: need to pass an argument to configure so that the processors are recreated
            self.tracer.configure(api_version="v0.4")
            payload = "1' or '1' = '1'"
            self.app.post(
                url_for(controller="root", action="body"),
                params=payload,
                extra_environ={"CONTENT_TYPE": "text/plain"},
            )

            spans = self.pop_spans()
            assert spans

            root_span = spans[0]
            appsec_json = root_span.get_tag("_dd.appsec.json")
            assert "triggers" in json.loads(appsec_json if appsec_json else "{}")

            span = _context.get_item("http.request.body", span=root_span)
            assert span
            assert span == "1' or '1' = '1'"

    def test_request_method_get_200(self):
        res = self.app.get(url_for(controller="root", action="index"))
        assert res.status == 200
        spans = self.pop_spans()
        assert spans[0].get_tag("http.method") == "GET"

    def test_request_method_get_404(self):
        with pytest.raises(Exception):
            res = self.app.get(url_for(controller="root", action="index") + "nonexistent-path")
            assert res.status == 404
        spans = self.pop_spans()
        assert spans[0].get_tag("http.method") == "GET"

    def test_request_method_post_200(self):
        res = self.app.post(url_for(controller="root", action="index"))
        assert res.status == 200
        spans = self.pop_spans()
        assert spans[0].get_tag("http.method") == "POST"

    def test_pylon_path_params(self):
        with override_global_config(dict(_appsec_enabled=True)):
            self.tracer._appsec_enabled = False
            self.tracer.configure(api_version="v0.4")
            self.app.get("/path-params/2022/july/")

            spans = self.pop_spans()
            root_span = spans[0]
            assert root_span.get_tag("_dd.appsec.json") is None
            path_params = _context.get_item("http.request.path_params", span=root_span)

            assert path_params["month"] == "july"
            assert path_params["year"] == "2022"

    def test_pylon_path_params_attack(self):
        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)):
            self.tracer._appsec_enabled = True

            self.tracer.configure(api_version="v0.4")
            self.app.get("/path-params/2022/w00tw00t.at.isc.sans.dfind/")

            spans = self.pop_spans()
            root_span = spans[0]

            appsec_json = root_span.get_tag("_dd.appsec.json")
            assert "triggers" in json.loads(appsec_json if appsec_json else "{}")

            query = dict(_context.get_item("http.request.path_params", span=root_span))
            assert query["month"] == "w00tw00t.at.isc.sans.dfind"
            assert query["year"] == "2022"

    def test_pylons_useragent(self):
        self.app.get(url_for(controller="root", action="index"), headers={"HTTP_USER_AGENT": "test/1.2.3"})
        spans = self.pop_spans()
        root_span = spans[0]
        assert root_span.get_tag(http.USER_AGENT) == "test/1.2.3"

    def test_pylons_body_json_empty_body(self):
        """
        "Failed to parse request body"
        """
        with self._caplog.at_level(logging.WARNING), override_global_config(dict(_appsec_enabled=True)):
            # Hack: need to pass an argument to configure so that the processors are recreated
            self.tracer.configure(api_version="v0.4")
            payload = ""

            self.app.post(
                url_for(controller="root", action="body"),
                params=payload,
                extra_environ={"CONTENT_TYPE": "application/json"},
            )
            assert "Failed to parse request body" in self._caplog.text

    def test_pylon_get_user(self):
        self.app.get("/identify")

        spans = self.pop_spans()
        root_span = spans[0]

        # Values defined in tests/contrib/pylons/app/controllers/root.py::RootController::identify
        assert root_span.get_tag(user.ID) == "usr.id"
        assert root_span.get_tag(user.EMAIL) == "usr.email"
        assert root_span.get_tag(user.SESSION_ID) == "usr.session_id"
        assert root_span.get_tag(user.NAME) == "usr.name"
        assert root_span.get_tag(user.ROLE) == "usr.role"
        assert root_span.get_tag(user.SCOPE) == "usr.scope"



import os
from typing import Generator

import pytest

from ddtrace import Tracer
from ddtrace.constants import SPAN_MEASURED_KEY
from ddtrace.ext import http
from ddtrace.internal.processor.stats import SpanStatsProcessorV06
from ddtrace.sampler import DatadogSampler
from ddtrace.sampler import SamplingRule
from tests.utils import override_global_config

from .test_integration import AGENT_VERSION


pytestmark = pytest.mark.skipif(AGENT_VERSION != "testagent", reason="Tests only compatible with a testagent")


@pytest.fixture
def sample_rate():
    # type: () -> Generator[float, None, None]
    # Default the sample rate to 0 so no traces are sent for requests.
    yield 0.0


@pytest.fixture
def stats_tracer(sample_rate):
    # type: (float) -> Generator[Tracer, None, None]
    with override_global_config(dict(_trace_compute_stats=True)):
        tracer = Tracer()
        tracer.configure(
            sampler=DatadogSampler(
                default_sample_rate=sample_rate,
            )
        )
        yield tracer
        tracer.shutdown()


@pytest.mark.parametrize("envvar", ["DD_TRACE_STATS_COMPUTATION_ENABLED", "DD_TRACE_COMPUTE_STATS"])
def test_compute_stats_default_and_configure(run_python_code_in_subprocess, envvar):
    """Ensure stats computation can be enabled."""

    # Test enabling via `configure`
    t = Tracer()
    assert not t._compute_stats
    assert not any(isinstance(p, SpanStatsProcessorV06) for p in t._span_processors)
    t.configure(compute_stats_enabled=True)
    assert any(isinstance(p, SpanStatsProcessorV06) for p in t._span_processors)
    assert t._compute_stats

    # Test enabling via environment variable
    env = os.environ.copy()
    env.update({envvar: "true"})
    out, err, status, _ = run_python_code_in_subprocess(
        """
from ddtrace import tracer
from ddtrace import config
from ddtrace.internal.processor.stats import SpanStatsProcessorV06
assert config._trace_compute_stats is True
assert any(isinstance(p, SpanStatsProcessorV06) for p in tracer._span_processors)
""",
        env=env,
    )
    assert status == 0, out + err


# Can't use a value between 0 and 1 since sampling is not deterministic.
@pytest.mark.parametrize("sample_rate", [1.0, 0.0])
@pytest.mark.snapshot()
def test_sampling_rate(stats_tracer, sample_rate):
    """Ensure traces are sent according to the sampling rate."""
    for _ in range(10):
        with stats_tracer.trace("operation"):
            pass


@pytest.mark.snapshot()
def test_stats_1000(stats_tracer, sample_rate):
    for i in range(1000):
        with stats_tracer.trace("name", service="abc", resource="/users/list"):
            pass


@pytest.mark.snapshot()
def test_stats_errors(stats_tracer, sample_rate):
    for i in range(1000):
        with stats_tracer.trace("name", service="abc", resource="/users/list") as span:
            if i % 2 == 0:
                span.error = 1


@pytest.mark.snapshot()
def test_stats_aggrs(stats_tracer, sample_rate):
    """
    When different span properties are set
        The stats are put into different aggregations
    """
    with stats_tracer.trace(name="op", service="my-svc", span_type="web", resource="/users/list"):
        pass

    # Synthetics
    with stats_tracer.trace(name="op", service="my-svc", span_type="web", resource="/users/list") as span:
        span.context.dd_origin = "synthetics"

    # HTTP status code
    with stats_tracer.trace(name="op", service="my-svc", span_type="web", resource="/users/list") as span:
        span.set_tag(http.STATUS_CODE, 200)

    # Resource
    with stats_tracer.trace(name="op", service="my-svc", span_type="web", resource="/users/view"):
        pass

    # Service
    with stats_tracer.trace(name="op", service="diff-svc", span_type="web", resource="/users/list"):
        pass

    # Name
    with stats_tracer.trace(name="diff-op", service="my-svc", span_type="web", resource="/users/list"):
        pass

    # Type
    with stats_tracer.trace(name="diff-op", service="my-svc", span_type="db", resource="/users/list"):
        pass


@pytest.mark.snapshot()
def test_measured_span(stats_tracer):
    for _ in range(10):
        with stats_tracer.trace("parent"):  # Should have stats
            with stats_tracer.trace("child"):  # Shouldn't have stats
                pass
    for _ in range(10):
        with stats_tracer.trace("parent"):  # Should have stats
            with stats_tracer.trace("child_stats") as span:  # Should have stats
                span.set_tag(SPAN_MEASURED_KEY)


@pytest.mark.snapshot()
def test_top_level(stats_tracer):
    for _ in range(100):
        with stats_tracer.trace("parent", service="svc-one"):  # Should have stats
            with stats_tracer.trace("child", service="svc-two"):  # Should have stats
                pass


@pytest.mark.parametrize(
    "sampling_rule",
    [SamplingRule(sample_rate=0, service="test"), SamplingRule(sample_rate=1, service="test")],
)
@pytest.mark.snapshot()
def test_single_span_sampling(stats_tracer, sampling_rule):
    sampler = DatadogSampler([sampling_rule])
    stats_tracer.configure(sampler=sampler)
    with stats_tracer.trace("parent", service="test"):
        with stats_tracer.trace("child") as child:
            # FIXME: Replace with span sampling rule
            child.set_metric("_dd.span_sampling.mechanism", 8)



from celery import registry

from ddtrace import Pin
from ddtrace import config

from . import constants as c
from .. import trace_utils
from ...constants import ANALYTICS_SAMPLE_RATE_KEY
from ...constants import SPAN_MEASURED_KEY
from ...ext import SpanTypes
from ...internal.logger import get_logger
from ...propagation.http import HTTPPropagator
from .utils import attach_span
from .utils import detach_span
from .utils import retrieve_span
from .utils import retrieve_task_id
from .utils import set_tags_from_context


log = get_logger(__name__)
propagator = HTTPPropagator


def trace_prerun(*args, **kwargs):
    # safe-guard to avoid crashes in case the signals API
    # changes in Celery
    task = kwargs.get("sender")
    task_id = kwargs.get("task_id")
    log.debug("prerun signal start task_id=%s", task_id)
    if task is None or task_id is None:
        log.debug("unable to extract the Task and the task_id. This version of Celery may not be supported.")
        return

    # retrieve the task Pin or fallback to the global one
    pin = Pin.get_from(task) or Pin.get_from(task.app)
    if pin is None:
        log.debug("no pin found on task or task.app task_id=%s", task_id)
        return

    request_headers = task.request.get("headers", {})
    trace_utils.activate_distributed_headers(pin.tracer, int_config=config.celery, request_headers=request_headers)

    # propagate the `Span` in the current task Context
    service = config.celery["worker_service_name"]
    span = pin.tracer.trace(c.WORKER_ROOT_SPAN, service=service, resource=task.name, span_type=SpanTypes.WORKER)
    # set analytics sample rate
    rate = config.celery.get_analytics_sample_rate()
    if rate is not None:
        span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, rate)

    span.set_tag(SPAN_MEASURED_KEY)
    attach_span(task, task_id, span)


def trace_postrun(*args, **kwargs):
    # safe-guard to avoid crashes in case the signals API
    # changes in Celery
    task = kwargs.get("sender")
    task_id = kwargs.get("task_id")
    log.debug("postrun signal task_id=%s", task_id)
    if task is None or task_id is None:
        log.debug("unable to extract the Task and the task_id. This version of Celery may not be supported.")
        return

    # retrieve and finish the Span
    span = retrieve_span(task, task_id)
    if span is None:
        log.warning("no existing span found for task_id=%s", task_id)
        return
    else:
        # request context tags
        span.set_tag(c.TASK_TAG_KEY, c.TASK_RUN)
        set_tags_from_context(span, kwargs)
        set_tags_from_context(span, task.request.__dict__)
        span.finish()
        detach_span(task, task_id)


def trace_before_publish(*args, **kwargs):
    # `before_task_publish` signal doesn't propagate the task instance so
    # we need to retrieve it from the Celery Registry to access the `Pin`. The
    # `Task` instance **does not** include any information about the current
    # execution, so it **must not** be used to retrieve `request` data.
    task_name = kwargs.get("sender")
    task = registry.tasks.get(task_name)
    task_id = retrieve_task_id(kwargs)
    # safe-guard to avoid crashes in case the signals API
    # changes in Celery
    if task is None or task_id is None:
        log.debug("unable to extract the Task and the task_id. This version of Celery may not be supported.")
        return

    # propagate the `Span` in the current task Context
    pin = Pin.get_from(task) or Pin.get_from(task.app)
    if pin is None:
        return

    # apply some tags here because most of the data is not available
    # in the task_after_publish signal
    service = config.celery["producer_service_name"]
    span = pin.tracer.trace(c.PRODUCER_ROOT_SPAN, service=service, resource=task_name)
    # set analytics sample rate
    rate = config.celery.get_analytics_sample_rate()
    if rate is not None:
        span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, rate)

    span.set_tag(SPAN_MEASURED_KEY)
    span.set_tag(c.TASK_TAG_KEY, c.TASK_APPLY_ASYNC)
    span.set_tag("celery.id", task_id)
    set_tags_from_context(span, kwargs)

    # Note: adding tags from `traceback` or `state` calls will make an
    # API call to the backend for the properties so we should rely
    # only on the given `Context`
    attach_span(task, task_id, span, is_publish=True)

    if config.celery["distributed_tracing"]:
        trace_headers = {}
        propagator.inject(span.context, trace_headers)

        # This weirdness is due to yet another Celery bug concerning
        # how headers get propagated in async flows
        # https://github.com/celery/celery/issues/4875
        task_headers = kwargs.get("headers") or {}
        task_headers.setdefault("headers", {})
        task_headers["headers"].update(trace_headers)
        kwargs["headers"] = task_headers


def trace_after_publish(*args, **kwargs):
    task_name = kwargs.get("sender")
    task = registry.tasks.get(task_name)
    task_id = retrieve_task_id(kwargs)
    # safe-guard to avoid crashes in case the signals API
    # changes in Celery
    if task is None or task_id is None:
        log.debug("unable to extract the Task and the task_id. This version of Celery may not be supported.")
        return

    # retrieve and finish the Span
    span = retrieve_span(task, task_id, is_publish=True)
    if span is None:
        return
    else:
        span.finish()
        detach_span(task, task_id, is_publish=True)


def trace_failure(*args, **kwargs):
    # safe-guard to avoid crashes in case the signals API
    # changes in Celery
    task = kwargs.get("sender")
    task_id = kwargs.get("task_id")
    if task is None or task_id is None:
        log.debug("unable to extract the Task and the task_id. This version of Celery may not be supported.")
        return

    # retrieve and finish the Span
    span = retrieve_span(task, task_id)
    if span is None:
        return
    else:
        # add Exception tags; post signals are still called
        # so we don't need to attach other tags here
        ex = kwargs.get("einfo")
        if ex is None:
            return
        if hasattr(task, "throws") and isinstance(ex.exception, task.throws):
            return
        span.set_exc_info(ex.type, ex.exception, ex.tb)


def trace_retry(*args, **kwargs):
    # safe-guard to avoid crashes in case the signals API
    # changes in Celery
    task = kwargs.get("sender")
    context = kwargs.get("request")
    if task is None or context is None:
        log.debug("unable to extract the Task or the Context. This version of Celery may not be supported.")
        return

    reason = kwargs.get("reason")
    if not reason:
        log.debug("unable to extract the retry reason. This version of Celery may not be supported.")
        return

    span = retrieve_span(task, context.id)
    if span is None:
        return

    # Add retry reason metadata to span
    # DEV: Use `str(reason)` instead of `reason.message` in case we get something that isn't an `Exception`
    span.set_tag(c.TASK_RETRY_REASON_KEY, str(reason))
